#!/usr/bin/env ruby

require 'date'
require "optparse"
require "ostruct"
require "xmlrpc/client"

# A couple things that still need to be done:
#   - Add a quiet mode
#   - Option for the event builders to sprout a canvas
#   - Make disk writing optional
#   - Validate configurations looking for simple problems
#   - Calculate the optimum number of buffers for each process

# It'd be nice if we were using a newer version of ruby (1.9.3) that had
# better string format substitution.  For the time being we'll setup our
# string constants using the newer convention and do the ugly substitutions.

# 17-Sep-2013, KAB - provide a way to fetch the online monitoring
# configuration from a separate file
if ENV['DS50DAQ_CONFIG_PATH']
  @cfgPathArray = ENV['DS50DAQ_CONFIG_PATH'].split(':')
  @cfgPathArray = @cfgPathArray.reverse
  (@cfgPathArray).each do |path|
    $LOAD_PATH.unshift path
  end
end
begin
  require 'onmon_config'
rescue Exception => msg
end

# Please NOTE that this assignment for the ONMON_CFG will only
# be used if it wasn't included from the external file already
if (defined?(PHYS_ANAL_ONMON_CFG)).nil? || (PHYS_ANAL_ONMON_CFG).nil?
  PHYS_ANAL_ONMON_CFG = "\
    app: {
      module_type: RootApplication
      force_new: true
    }
    wf: {
      module_type: WFViewer
      prescale: 20
      digital_sum_only: false
    }
    baseline: {
      module_type: MonitorBaseline
      random_only: false
    }
    trigger: {
      module_type: MonitorTrigger
    }
    ser: {
      module_type: MonitorSer
      start_us: 0.100
      stop_us: 0.180
      laser_only: true
      #single_channel: 0
      range: [ -1200, 100 ]
    }"
end
# And, this assignment for the prescale will only
# be used if it wasn't included from the external file already
if (defined?(ONMON_EVENT_PRESCALE)).nil? || (ONMON_EVENT_PRESCALE).nil?
  ONMON_EVENT_PRESCALE = 1
end
# ditto, the online monitoring modules that are run
if (defined?(ONMON_MODULES)).nil? || (ONMON_MODULES).nil?
  ONMON_MODULES = "[ app, wf, trigger, ser ]"
end
# ditto, the online monitoring prolog
if (defined?(ONMON_PROLOG)).nil? || (ONMON_PROLOG).nil?
  ONMON_PROLOG = ""
end

EB_CONFIG = "\
BEGIN_PROLOG
  huff:
  {
    module_type: DS50Compression
    raw_label: daq
    want_bins: false
    perf_print: false
    use_diffs: true
    record_compression: false
  }

  huffdiff_1720: @local::huff
  huffdiff_1720.instance_name: V1720
  huffdiff_1720.use_db_for_table: %{use_db_for_table}
  %{symtable_name}huffdiff_1720.symtable_name: \"v1720_huff_diff\"
  %{table_file}huffdiff_1720.table_file: \"table_3544_daqV1720_huff_diff.txt\"
  huffdiff_1724: @local::huff
  huffdiff_1724.instance_name: V1724
  huffdiff_1724.use_db_for_table: %{use_db_for_table}
  %{symtable_name}huffdiff_1724.symtable_name: \"v1724_huff_diff\"
  %{table_file}huffdiff_1724.table_file: \"table_daqV1724_huff_diff.txt\"
END_PROLOG
services: {
  scheduler: {
    fileMode: NOMERGE
  }
  user: {
    NetMonTransportServiceInterface: {
      service_provider: NetMonTransportService
      first_data_receiver_rank: %{ag_rank}
      mpi_buffer_count: %{netmonout_buffer_count}
      max_fragment_size_words: %{size_words}
      data_receiver_count: 1 # %{ag_count}
      #broadcast_sends: true
    }
    DBInterface: {
      hostname: \"dsdb.lngs.infn.it\"
      database: \"darkside50\"
      user: \"rea\"
      password: \"der\"
    }
  }
  Timing: { summaryOnly: true }
  #SimpleMemoryCheck: { }
}
daq: {
  max_fragment_size_words: %{size_words}
  event_builder: {
    mpi_buffer_count: %{buffer_count}
    first_fragment_receiver_rank: 0
    fragment_receiver_count: %{total_frs}
    expected_fragments_per_event: %{total_fragments}
    use_art: true
    print_event_store_stats: true
    event_queue_depth: 20
    event_queue_wait_time: 10.0
  }
}
outputs: {
  %{netmon_output}netMonOutput: {
  %{netmon_output}  module_type: NetMonOutput
  %{netmon_output}  %{drop_uncompressed}outputCommands: [ \"keep *\", \"drop artdaq::Fragments_daq_V1720_*\", \"drop artdaq::Fragments_daq_V1724_*\" ]
  %{netmon_output}}
  %{root_output}normalOutput: {
  %{root_output}  module_type: RootOutput
  %{root_output}  fileName: \"%{output_file}\"
  %{root_output}  compressionLevel: 0
  %{root_output}  %{drop_uncompressed}outputCommands: [ \"keep *\", \"drop artdaq::Fragments_daq_V1720_*\", \"drop artdaq::Fragments_daq_V1724_*\" ]
  %{root_output}}
}

physics: {
  analyzers: {
%{phys_anal_onmon_cfg}
  }

  producers: {
    # compress the raw
    %{enable_1720_compression}huffdiffV1720: @local::huffdiff_1720  
    %{enable_1724_compression}huffdiffV1724: @local::huffdiff_1724 
    %{enable_172x_compression}huffdiffV1720: @local::huffdiff_1720  
    %{enable_172x_compression}huffdiffV1724: @local::huffdiff_1724 
  }
  %{enable_1720_compression}p1: [ huffdiffV1720 ]
  %{enable_1724_compression}p1: [ huffdiffV1724 ]
  %{enable_172x_compression}p1: [ huffdiffV1720, huffdiffV1724 ]

  %{enable_onmon}a1: [ app, wf, trigger, baseline, ser ]

  %{netmon_output}my_output_modules: [ netMonOutput ]
  %{root_output}my_output_modules: [ normalOutput ]
}
source: {
  module_type: RawInput
  waiting_time: 900
  resume_after_timeout: true
  fragment_type_map: [[1, \"missed\"], [2, \"V1495\"], [3, \"V1720\"], [4, \"V1724\"], [5, \"V1190\"], [6, \"V1495Veto\"]]
}
process_name: DAQ"

AG_CONFIG = "\
BEGIN_PROLOG
  huff:
  {
    module_type: DS50Compression
    raw_label: daq
    want_bins: false
    perf_print: false
    use_diffs: true
    record_compression: false
  }

  huffdiff_1720: @local::huff
  huffdiff_1720.instance_name: V1720
  huffdiff_1720.use_db_for_table: %{use_db_for_table}
  %{symtable_name}huffdiff_1720.symtable_name: \"v1720_huff_diff\"
  %{table_file}huffdiff_1720.table_file: \"table_3544_daqV1720_huff_diff.txt\"
  huffdiff_1724: @local::huff
  huffdiff_1724.instance_name: V1724
  huffdiff_1724.use_db_for_table: %{use_db_for_table}
  %{symtable_name}huffdiff_1724.symtable_name: \"v1724_huff_diff\"
  %{table_file}huffdiff_1724.table_file: \"table_daqV1724_huff_diff.txt\"

  unhuff:
  {
    module_type: DS50Decompression
  }

  unhuff_1720: @local::unhuff
  unhuff_1720.instance_name: V1720
  unhuff_1720.compressed_label: huffdiffV1720
  unhuff_1720.use_db_for_table: %{use_db_for_table}
  %{symtable_name}unhuff_1720.symtable_name: \"v1720_huff_diff\"
  %{table_file}unhuff_1720.table_file: \"table_3544_daqV1720_huff_diff.txt\"
  unhuff_1724: @local::unhuff
  unhuff_1724.instance_name: V1724
  unhuff_1724.compressed_label: huffdiffV1724
  unhuff_1724.use_db_for_table: %{use_db_for_table}
  %{symtable_name}unhuff_1724.symtable_name: \"v1724_huff_diff\"
  %{table_file}unhuff_1724.table_file: \"table_daqV1724_huff_diff.txt\"
  %{onmon_prolog}
END_PROLOG
services: {
  scheduler: {
    fileMode: NOMERGE
  }
  user: {
    NetMonTransportServiceInterface: {
      service_provider: NetMonTransportService
      max_fragment_size_words: %{size_words}
    }
    DBInterface: {
      hostname: \"dsdb.lngs.infn.it\"
      database: \"darkside50\"
      user: \"rea\"
      password: \"der\"
    }
  }
  Timing: { summaryOnly: true }
  #SimpleMemoryCheck: { }
}
daq: {
  max_fragment_size_words: %{size_words}
  aggregator: {
    mpi_buffer_count: %{buffer_count}
    first_event_builder_rank: %{total_frs}
    event_builder_count: %{total_ebs}
    expected_events_per_bunch: %{bunch_size}
    print_event_store_stats: true
    event_queue_depth: %{queue_depth}
    event_queue_wait_time: %{queue_timeout}
    onmon_event_prescale: %{onmon_event_prescale}
    xmlrpc_client_list: \"%{xmlrpc_client_list}\"
    file_size_MB: %{file_size}
    file_duration: %{file_duration}
    file_event_count: %{file_event_count}
  }
}
source: {
  module_type: NetMonInput
}
outputs: {
  %{root_output}normalOutput: {
  %{root_output}  module_type: RootOutput
  %{root_output}  fileName: \"%{output_file}\"
  %{root_output}  compressionLevel: 0
  %{root_output}  %{drop_uncompressed}outputCommands: [ \"keep *\", \"drop artdaq::Fragments_daq_V1720_*\", \"drop artdaq::Fragments_daq_V1724_*\" ]
  %{root_output}}
}
physics: {
  analyzers: {
%{phys_anal_onmon_cfg}
  }

  producers: {
    # compress the raw
    %{enable_1720_compression}huffdiffV1720: @local::huffdiff_1720  
    %{enable_1724_compression}huffdiffV1724: @local::huffdiff_1724 
    %{enable_172x_compression}huffdiffV1720: @local::huffdiff_1720  
    %{enable_172x_compression}huffdiffV1724: @local::huffdiff_1724 

    ## decompress the compressed
    %{enable_1720_decompression}unhuffV1720: @local::unhuff_1720  
    %{enable_1724_decompression}unhuffV1724: @local::unhuff_1724 
    %{enable_172x_decompression}unhuffV1720: @local::unhuff_1720  
    %{enable_172x_decompression}unhuffV1724: @local::unhuff_1724 

    BuildInfo:
    {
      module_type: DS50DaqBuildInfo
      instance_name: ds50daq
    }
  }
  %{enable_1720_compression}p1: [ huffdiffV1720 ]
  %{enable_1724_compression}p1: [ huffdiffV1724 ]
  %{enable_172x_compression}p1: [ huffdiffV1720, huffdiffV1724 ]

  %{enable_1720_decompression}p2: [ unhuffV1720 ]
  %{enable_1724_decompression}p2: [ unhuffV1724 ]
  %{enable_172x_decompression}p2: [ unhuffV1720, unhuffV1724 ]

  %{enable_onmon}a1: %{onmon_modules}

  p5: [ BuildInfo ]

  %{root_output}my_output_modules: [ normalOutput ]
}
process_name: DAQAG"

COMPOSITE_GENERATOR_CONFIG = "\
%{prolog}
daq: {
  max_fragment_size_words: %{size_words}
  fragment_receiver: {
    mpi_buffer_count: %{buffer_count}
    first_event_builder_rank: %{total_frs}
    event_builder_count: %{total_ebs}
    generator: CompositeDriver
    fragment_id: 999
    board_id: 999
    generator_config_list:
    [
      # the format of this list is {daq:<paramSet},{daq:<paramSet>},...
      %{generator_list}
    ]
  }
}"

V1495_CONFIG = "\
daq: {
  max_fragment_size_words: %{size_words}
  fragment_receiver: {
    mpi_buffer_count: %{buffer_count}
    first_event_builder_rank: %{total_frs}
    event_builder_count: %{total_ebs}
    generator: V1495Driver
    masked_channels_0bias: [ 38, 39 ]
    link_type: PCIE
    link_number: %{link_number}
    vme_base: 0x01000000
    pulser_frequency: %{pulser_rate}
    laser_frequency: %{laser_rate}
    random_triggers: %{random_trigger_enable}
    acquisition_gate_us: %{gate_width}
    inhibit_us: %{inhibit_time}
    low_threshold: %{low_threshold}
    high_threshold: 39
    majority_time_10ns: 10
    enable_veto_busy: %{enable_veto_busy}
    g2_mode: %{g2_mode}
    g2_gate_us: 5
    g2_prescale: 33
    g2_threshold: 360
    g2_threshold_high: 1500
    tdc_pulse_delay_us: %{tdc_pulse_delay}
    %{fnal_mode}fnal_mode: true
    generator_ds50: {
      fragment_id: %{fragment_id}
      board_id: %{board_id}
    }
    rt_priority: 2
    # RC_PLACEHOLDER
  }
}"

VETO_V1495_CONFIG = "\
daq: {
  max_fragment_size_words: %{size_words}
  fragment_receiver: {
    mpi_buffer_count: %{buffer_count}
    first_event_builder_rank: %{total_frs}
    event_builder_count: %{total_ebs}
    generator: V1495Driver
    link_type: PCIE
    link_number: %{link_number}
    vme_base: 0x01000000
    acquisition_gate_us: %{gate_width}
    tdc_pulse_delay_us: %{tdc_pulse_delay}
    veto: true
    global: false
    fanout_vme_base: %{fanout_vme_base}
    veto_choice: %{veto_choice}
    enable_tdc_busy: %{enable_tdc_busy}
    generator_ds50: {
      fragment_id: %{fragment_id}
      board_id: %{board_id}
    }
    rt_priority: 2
    # RC_PLACEHOLDER
  }
}"

V1720_CONFIG = "\
BEGIN_PROLOG
def_offset: 0x2700
def_channels: [ { channel: 0 enable: true dc_offset: @local::def_offset },
                { channel: 1 enable: true dc_offset: @local::def_offset },
                { channel: 2 enable: true dc_offset: @local::def_offset },
                { channel: 3 enable: true dc_offset: @local::def_offset },
                { channel: 4 enable: true dc_offset: @local::def_offset },
                { channel: 5 enable: true dc_offset: @local::def_offset },
                { channel: 6 enable: true dc_offset: @local::def_offset },
                { channel: 7 enable: true dc_offset: @local::def_offset } ]
END_PROLOG
daq: {
  max_fragment_size_words: %{size_words}
  fragment_receiver: {
    mpi_buffer_count: %{buffer_count}
    first_event_builder_rank: %{total_frs}
    event_builder_count: %{total_ebs}
    generator: V172xDriver
    link_type: PCIE
    link_number: %{link_number}
    io_level: TTL
    acquisition_gate_us: %{gate_width}
    pre_trigger_us: %{pre_trigger}
    acquisition_mode: HW
    multiboard_sincronicazion: true
    #max_buffers_code: 4
    channels: @local::def_channels
    generator_ds50: {
      fragment_id: %{fragment_id}
      board_id: %{board_id}
    }
    rt_priority: 2
    # RC_PLACEHOLDER
  }
}"

V1720_SIM_CONFIG = "\
daq: {
  max_fragment_size_words: %{size_words}
  fragment_receiver: {
    mpi_buffer_count: %{buffer_count}
    mpi_sync_interval: 50
    first_event_builder_rank: %{total_frs}
    event_builder_count: %{total_ebs}
    generator: V172xSimulator 
    freqs_file: \"ds50_hist.dat\" 
    fragments_per_event: 1 
    nChannels: 80000 
    generator_ds50: {
      fragment_id: %{fragment_id}
      board_id: %{board_id}
    }
    rt_priority: 2
    # RC_PLACEHOLDER
  }
}"

V1190_CONFIG = "\
daq: {
  max_fragment_size_words: %{size_words}
  fragment_receiver: {
    mpi_buffer_count: %{buffer_count}
    first_event_builder_rank: %{total_frs}
    event_builder_count: %{total_ebs}
    generator: V1190Driver
    link_type: PCIE
    link_number: %{link_number}
    vme_base: %{vme_base}
    acquisition_gate_us: %{gate_width}
    post_trigger_us: %{post_trigger}
    irq_level: %{irq_level}
    irq_vector: %{irq_vector}
    generator_ds50: {
      fragment_id: %{fragment_id}
      board_id: %{board_id}
    }
    rt_priority: 2
    # RC_PLACEHOLDER
  }
}"

class ConfigGen
  def generateEventBuilder(ebIndex, totalFRs, totalEBs, totalAGs, compressionLevel, 
                           totalv1720s, totalv1724s, dataDir, onmonEnable,
                           diskWritingEnable, fragSizeWords, totalFragments,
                           compressionTableFromDB)
    # Do the substitutions in the event builder configuration given the options
    # that were passed in from the command line.  
    ebConfig = String.new(EB_CONFIG)
    ebConfig.gsub!(/\%\{total_frs\}/, String(totalFRs))
    ebConfig.gsub!(/\%\{total_fragments\}/, String(totalFragments))
    ebConfig.gsub!(/\%\{ag_rank\}/, String(totalFRs + totalEBs))
    ebConfig.gsub!(/\%\{ag_count\}/, String(totalAGs))
    ebConfig.gsub!(/\%\{buffer_count\}/, String(totalFRs*8))
    ebConfig.gsub!(/\%\{size_words\}/, String(fragSizeWords))
    ebConfig.gsub!(/\%\{netmonout_buffer_count\}/, String(totalAGs*4))
    if Integer(compressionLevel) > 10 && Integer(compressionLevel) < 13
      if Integer(totalv1720s) > 0 and Integer(totalv1724s) > 0
        ebConfig.gsub!(/\%\{enable_1720_compression\}/, "#")
        ebConfig.gsub!(/\%\{enable_1724_compression\}/, "#")
        ebConfig.gsub!(/\%\{enable_172x_compression\}/, "")
      elsif Integer(totalv1720s) > 0
        ebConfig.gsub!(/\%\{enable_1720_compression\}/, "")
        ebConfig.gsub!(/\%\{enable_1724_compression\}/, "#")
        ebConfig.gsub!(/\%\{enable_172x_compression\}/, "#")
      elsif Integer(totalv1724s) > 0
        ebConfig.gsub!(/\%\{enable_1720_compression\}/, "#")
        ebConfig.gsub!(/\%\{enable_1724_compression\}/, "")
        ebConfig.gsub!(/\%\{enable_172x_compression\}/, "#")
      else
        ebConfig.gsub!(/\%\{enable_1720_compression\}/, "#")
        ebConfig.gsub!(/\%\{enable_1724_compression\}/, "#")
        ebConfig.gsub!(/\%\{enable_172x_compression\}/, "#")
      end
    else
      ebConfig.gsub!(/\%\{enable_1720_compression\}/, "#")
      ebConfig.gsub!(/\%\{enable_1724_compression\}/, "#")
      ebConfig.gsub!(/\%\{enable_172x_compression\}/, "#")
    end
    if Integer(compressionLevel) > 11 && Integer(compressionLevel) < 14
      ebConfig.gsub!(/\%\{drop_uncompressed\}/, "")
    else
      ebConfig.gsub!(/\%\{drop_uncompressed\}/, "#")
    end
    if Integer(totalAGs) >= 1
      ebConfig.gsub!(/\%\{netmon_output\}/, "")
      ebConfig.gsub!(/\%\{root_output\}/, "#")
      ebConfig.gsub!(/\%\{enable_onmon\}/, "#")
      ebConfig.gsub!(/\%\{phys_anal_onmon_cfg\}/, "")
    else
      ebConfig.gsub!(/\%\{netmon_output\}/, "#")
      if Integer(diskWritingEnable) != 0
        ebConfig.gsub!(/\%\{root_output\}/, "")
      else
        ebConfig.gsub!(/\%\{root_output\}/, "#")
      end
      if Integer(onmonEnable) != 0
        ebConfig.gsub!(/\%\{phys_anal_onmon_cfg\}/, PHYS_ANAL_ONMON_CFG)
        ebConfig.gsub!(/\%\{enable_onmon\}/, "")
      else
        ebConfig.gsub!(/\%\{phys_anal_onmon_cfg\}/, "")
        ebConfig.gsub!(/\%\{enable_onmon\}/, "#")
      end
    end
    if Integer(compressionTableFromDB) != 0
      ebConfig.gsub!(/\%\{use_db_for_table\}/, "true")
      ebConfig.gsub!(/\%\{symtable_name\}/, "")
      ebConfig.gsub!(/\%\{table_file\}/, "#")
    else
      ebConfig.gsub!(/\%\{use_db_for_table\}/, "false")
      ebConfig.gsub!(/\%\{symtable_name\}/, "#")
      ebConfig.gsub!(/\%\{table_file\}/, "")
    end

    currentTime = Time.now
    fileName = "ds50daq_eb%02d_" % ebIndex
    #fileName += "r%06r_sr%02s_%to"
    fileName += "%to"
    #fileName += "%d%02d%02d-%02d%02d%02d" % [currentTime.year, currentTime.month,
    #                                         currentTime.day, currentTime.hour,
    #                                         currentTime.min, currentTime.sec]
    fileName += ".root"
    outputFile = File.join(dataDir, fileName)
    ebConfig.gsub!(/\%\{output_file\}/, outputFile)

    return ebConfig
  end

  def generateAggregator(dataDir, runNumber, totalFRs, totalEBs, bunchSize,
                         compressionLevel, totalv1720s, totalv1724s, onmonEnable,
                         diskWritingEnable, agIndex, totalAGs, fragSizeWords,
                         xmlrpcClientList, fileSizeThreshold, fileDuration,
                         fileEventCount, compressionTableFromDB)
    puts "Initial aggregator " + String(agIndex) + " disk writing setting = " +
      String(diskWritingEnable)
    # Do the substitutions in the aggregator configuration given the options
    # that were passed in from the command line.  Assure that files written out
    # by each AG are unique by including a timestamp in the file name.
    currentTime = Time.now
    #fileName = "ds50daq_AG%02d_" % [Integer(agIndex)]
    fileName = "ds50daq_"
    #fileName += "r%06r_sr%02s_%to"
    fileName += "%to"
    #fileName += "%d%02d%02d-%02d%02d%02d" % [currentTime.year, currentTime.month,
    #                                         currentTime.day, currentTime.hour,
    #                                         currentTime.min, currentTime.sec]
    fileName += ".root"
    outputFile = File.join(dataDir, fileName)
    agConfig = String.new(AG_CONFIG)
    agConfig.gsub!(/\%\{output_file\}/, outputFile)
    agConfig.gsub!(/\%\{total_ebs\}/, String(totalEBs))
    agConfig.gsub!(/\%\{total_frs\}/, String(totalFRs))
    agConfig.gsub!(/\%\{buffer_count\}/, String(totalEBs*4))
    agConfig.gsub!(/\%\{bunch_size\}/, String(bunchSize))
    agConfig.gsub!(/\%\{onmon_event_prescale\}/, String(ONMON_EVENT_PRESCALE))
    agConfig.gsub!(/\%\{onmon_modules\}/, String(ONMON_MODULES))
    agConfig.gsub!(/\%\{onmon_prolog\}/, String(ONMON_PROLOG))
    agConfig.gsub!(/\%\{size_words\}/, String(fragSizeWords))
    agConfig.gsub!(/\%\{xmlrpc_client_list\}/, String(xmlrpcClientList))
    agConfig.gsub!(/\%\{file_size\}/, String(fileSizeThreshold))
    agConfig.gsub!(/\%\{file_duration\}/, String(fileDuration))
    agConfig.gsub!(/\%\{file_event_count\}/, String(fileEventCount))
    if Integer(compressionLevel) > 10 && Integer(compressionLevel) < 13
      if Integer(totalv1720s) > 0 and Integer(totalv1724s) > 0
        agConfig.gsub!(/\%\{enable_1720_compression\}/, "#")
        agConfig.gsub!(/\%\{enable_1724_compression\}/, "#")
        agConfig.gsub!(/\%\{enable_172x_compression\}/, "")
      elsif Integer(totalv1720s) > 0
        agConfig.gsub!(/\%\{enable_1720_compression\}/, "")
        agConfig.gsub!(/\%\{enable_1724_compression\}/, "#")
        agConfig.gsub!(/\%\{enable_172x_compression\}/, "#")
      elsif Integer(totalv1724s) > 0
        agConfig.gsub!(/\%\{enable_1720_compression\}/, "#")
        agConfig.gsub!(/\%\{enable_1724_compression\}/, "")
        agConfig.gsub!(/\%\{enable_172x_compression\}/, "#")
      else
        agConfig.gsub!(/\%\{enable_1720_compression\}/, "#")
        agConfig.gsub!(/\%\{enable_1724_compression\}/, "#")
        agConfig.gsub!(/\%\{enable_172x_compression\}/, "#")
      end
    else
      agConfig.gsub!(/\%\{enable_1720_compression\}/, "#")
      agConfig.gsub!(/\%\{enable_1724_compression\}/, "#")
      agConfig.gsub!(/\%\{enable_172x_compression\}/, "#")
    end
    if Integer(compressionLevel) > 11 && Integer(compressionLevel) < 14
      agConfig.gsub!(/\%\{drop_uncompressed\}/, "")
    else
      agConfig.gsub!(/\%\{drop_uncompressed\}/, "#")
    end
    if Integer(compressionLevel) == 14
      if Integer(totalv1720s) > 0 and Integer(totalv1724s) > 0
        agConfig.gsub!(/\%\{enable_1720_decompression\}/, "#")
        agConfig.gsub!(/\%\{enable_1724_decompression\}/, "#")
        agConfig.gsub!(/\%\{enable_172x_decompression\}/, "")
      elsif Integer(totalv1720s) > 0
        agConfig.gsub!(/\%\{enable_1720_decompression\}/, "")
        agConfig.gsub!(/\%\{enable_1724_decompression\}/, "#")
        agConfig.gsub!(/\%\{enable_172x_decompression\}/, "#")
      elsif Integer(totalv1724s) > 0
        agConfig.gsub!(/\%\{enable_1720_decompression\}/, "#")
        agConfig.gsub!(/\%\{enable_1724_decompression\}/, "")
        agConfig.gsub!(/\%\{enable_172x_decompression\}/, "#")
      else
        agConfig.gsub!(/\%\{enable_1720_decompression\}/, "#")
        agConfig.gsub!(/\%\{enable_1724_decompression\}/, "#")
        agConfig.gsub!(/\%\{enable_172x_decompression\}/, "#")
      end
    else
      agConfig.gsub!(/\%\{enable_1720_decompression\}/, "#")
      agConfig.gsub!(/\%\{enable_1724_decompression\}/, "#")
      agConfig.gsub!(/\%\{enable_172x_decompression\}/, "#")
    end

    if agIndex == 0
      if totalAGs > 1
        onmonEnable = 0
      end
      agConfig.gsub!(/\%\{queue_depth\}/, "20")
      agConfig.gsub!(/\%\{queue_timeout\}/, "10.0")
    else
      diskWritingEnable = 0
      agConfig.gsub!(/\%\{queue_depth\}/, "2")
      agConfig.gsub!(/\%\{queue_timeout\}/, "1.0")
    end

    puts "Final aggregator " + String(agIndex) + " disk writing setting = " +
      String(diskWritingEnable)
    if Integer(diskWritingEnable) != 0
      agConfig.gsub!(/\%\{root_output\}/, "")
    else
      agConfig.gsub!(/\%\{root_output\}/, "#")
    end
    if Integer(onmonEnable) != 0
      agConfig.gsub!(/\%\{phys_anal_onmon_cfg\}/, PHYS_ANAL_ONMON_CFG)
      agConfig.gsub!(/\%\{enable_onmon\}/, "")
    else
      agConfig.gsub!(/\%\{phys_anal_onmon_cfg\}/, "")
      agConfig.gsub!(/\%\{enable_onmon\}/, "#")
    end
    if Integer(compressionTableFromDB) != 0
      agConfig.gsub!(/\%\{use_db_for_table\}/, "true")
      agConfig.gsub!(/\%\{symtable_name\}/, "")
      agConfig.gsub!(/\%\{table_file\}/, "#")
    else
      agConfig.gsub!(/\%\{use_db_for_table\}/, "false")
      agConfig.gsub!(/\%\{symtable_name\}/, "#")
      agConfig.gsub!(/\%\{table_file\}/, "")
    end

    return agConfig  
  end

  def generateV1495(fragmentId, totalEBs, totalFRs, linkNumber, pulserRate, laserRate,
                    randomTriggerEnable, gateWidth, lowThreshold, fnalMode, vetoChoice,
                    fragSizeWords, boardId, g2Mode, inhibitTime, tdcPulseDelay)
    # Do the substitutions in the V1495 configuration given the options that 
    # were passed in from the command line.
    v1495Config = String.new(V1495_CONFIG)
    v1495Config.gsub!(/\%\{total_ebs\}/, String(totalEBs))
    v1495Config.gsub!(/\%\{total_frs\}/, String(totalFRs))
    v1495Config.gsub!(/\%\{fragment_id\}/, String(fragmentId))
    v1495Config.gsub!(/\%\{board_id\}/, String(boardId))
    v1495Config.gsub!(/\%\{link_number\}/, String(linkNumber))
    v1495Config.gsub!(/\%\{pulser_rate\}/, String(pulserRate))
    v1495Config.gsub!(/\%\{laser_rate\}/, String(laserRate))
    if Integer(randomTriggerEnable) > 0
      v1495Config.gsub!(/\%\{random_trigger_enable\}/, "true")
    else
      v1495Config.gsub!(/\%\{random_trigger_enable\}/, "false")
    end
    v1495Config.gsub!(/\%\{low_threshold\}/, String(lowThreshold))
    v1495Config.gsub!(/\%\{buffer_count\}/, String(totalEBs*8))
    v1495Config.gsub!(/\%\{gate_width\}/, String(gateWidth))
    v1495Config.gsub!(/\%\{tdc_pulse_delay\}/, String(tdcPulseDelay))
    v1495Config.gsub!(/\%\{inhibit_time\}/, String(inhibitTime))
    if Integer(g2Mode) > 0
      v1495Config.gsub!(/\%\{g2_mode\}/, "true")
    else
      v1495Config.gsub!(/\%\{g2_mode\}/, "false")
    end
    if Integer(fnalMode) > 0
      v1495Config.gsub!(/\%\{fnal_mode\}/, "")
    else
      v1495Config.gsub!(/\%\{fnal_mode\}/, "#")
    end
    if vetoChoice.include? "off"
      v1495Config.gsub!(/\%\{enable_veto_busy\}/, "false")
    else
      v1495Config.gsub!(/\%\{enable_veto_busy\}/, "true")
    end
    v1495Config.gsub!(/\%\{size_words\}/, String(fragSizeWords))
    return v1495Config
  end

  def generateV1495Veto(fragmentId, totalEBs, totalFRs, linkNumber, gateWidth,
                        tdcPulseDelay, fanoutVMEBase, vetoChoice, fragSizeWords,
                        boardId)
    v1495Config = String.new(VETO_V1495_CONFIG)
    v1495Config.gsub!(/\%\{total_ebs\}/, String(totalEBs))
    v1495Config.gsub!(/\%\{total_frs\}/, String(totalFRs))
    v1495Config.gsub!(/\%\{fragment_id\}/, String(fragmentId))
    v1495Config.gsub!(/\%\{board_id\}/, String(boardId))
    v1495Config.gsub!(/\%\{link_number\}/, String(linkNumber))
    v1495Config.gsub!(/\%\{buffer_count\}/, String(totalEBs*8))
    v1495Config.gsub!(/\%\{gate_width\}/, String(gateWidth))
    v1495Config.gsub!(/\%\{tdc_pulse_delay\}/, String(tdcPulseDelay))
    v1495Config.gsub!(/\%\{fanout_vme_base\}/, fanoutVMEBase)
    if vetoChoice.include? "ADC"
      v1495Config.gsub!(/\%\{veto_choice\}/, 1)
    elsif vetoChoice.upcase.include? "HYBRID"
      v1495Config.gsub!(/\%\{veto_choice\}/, 2)
    else
      v1495Config.gsub!(/\%\{veto_choice\}/, 0)
    end
    if vetoChoice.include? "TDC"
      v1495Config.gsub!(/\%\{enable_tdc_busy\}/, "true")
    else
      v1495Config.gsub!(/\%\{enable_tdc_busy\}/, "false")
    end
    v1495Config.gsub!(/\%\{size_words\}/, String(fragSizeWords))
    return v1495Config
  end
  
  def generateV1720(fragmentId, totalEBs, totalFRs, linkNumber, sim,
                    gateWidth, preTriggerWindow, fragSizeWords, boardId)
    # Do the substitutions in the V1720 configuration given the options that 
    # were passed in from the command line.
    if sim != 0
      v1720Config = String.new(V1720_SIM_CONFIG)
    else
      v1720Config = String.new(V1720_CONFIG)
    end
    
    v1720Config.gsub!(/\%\{total_ebs\}/, String(totalEBs))
    v1720Config.gsub!(/\%\{total_frs\}/, String(totalFRs))
    v1720Config.gsub!(/\%\{link_number\}/, String(linkNumber))
    v1720Config.gsub!(/\%\{fragment_id\}/, String(fragmentId))
    v1720Config.gsub!(/\%\{board_id\}/, String(boardId))
    v1720Config.gsub!(/\%\{gate_width\}/, String(gateWidth))
    v1720Config.gsub!(/\%\{pre_trigger\}/, String(preTriggerWindow))
    v1720Config.gsub!(/\%\{buffer_count\}/, String(totalEBs*8))
    v1720Config.gsub!(/\%\{size_words\}/, String(fragSizeWords))
    return v1720Config
  end
  
  def generateV1190(fragmentId, totalEBs, totalFRs, linkNumber, vmeBase,
                    tdcGateWidth, postTrigger, irq_level, irq_vector, fragSizeWords,
                    boardId)
    # Do the substitutions in the V1190 configuration given the options that 
    # were passed in from the command line.
    v1190Config = String.new(V1190_CONFIG)
    v1190Config.gsub!(/\%\{total_ebs\}/, String(totalEBs))
    v1190Config.gsub!(/\%\{total_frs\}/, String(totalFRs))
    v1190Config.gsub!(/\%\{link_number\}/, String(linkNumber))
    v1190Config.gsub!(/\%\{fragment_id\}/, String(fragmentId))
    v1190Config.gsub!(/\%\{board_id\}/, String(boardId))
    v1190Config.gsub!(/\%\{buffer_count\}/, String(totalEBs*8))
    v1190Config.gsub!(/\%\{vme_base\}/, vmeBase)
    v1190Config.gsub!(/\%\{gate_width\}/, String(tdcGateWidth))
    v1190Config.gsub!(/\%\{post_trigger\}/, String(postTrigger))
    v1190Config.gsub!(/\%\{irq_level\}/, String(irq_level))
    v1190Config.gsub!(/\%\{irq_vector\}/, String(irq_vector))
    v1190Config.gsub!(/\%\{size_words\}/, String(fragSizeWords))
    return v1190Config
  end

  def generateComposite(totalEBs, totalFRs, configStringArray)
    compositeConfig = String.new(COMPOSITE_GENERATOR_CONFIG)
    compositeConfig.gsub!(/\%\{total_ebs\}/, String(totalEBs))
    compositeConfig.gsub!(/\%\{total_frs\}/, String(totalFRs))
    compositeConfig.gsub!(/\%\{buffer_count\}/, String(totalEBs*8))

    # The complications here are A) determining the largest buffer size that
    # has been requested by the child configurations and using that for the
    # composite buffers size and B) moving any PROLOG declarations from the
    # individual configuration strings to the front of the full CFG string.
    prologList = []
    fragSizeWords = 0
    configString = ""
    first = true
    configStringArray.each do |cfg|
      my_match = /(.*)BEGIN_PROLOG(.*)END_PROLOG(.*)/im.match(cfg)
      if my_match
        thisProlog = my_match[2]
        cfg = my_match[1] + my_match[3]

        found = false
        prologList.each do |savedProlog|
          if thisProlog == savedProlog
            found = true
            break
          end
        end
        if ! found
          prologList << thisProlog
        end
      end

      if first
        first = false
      else
        configString += ", "
      end
      configString += "{" + cfg + "}"

      my_match = /max_fragment_size_words\s*\:\s*(\d+)/.match(cfg)
      if my_match
        begin
          sizeWords = Integer(my_match[1])
          if sizeWords > fragSizeWords
            fragSizeWords = sizeWords
          end
        rescue Exception => msg
          puts "Warning: exception parsing size_words in composite child configuration: " + my_match[1] + " " + msg
        end
      end
    end
    compositeConfig.gsub!(/\%\{size_words\}/, String(fragSizeWords))
    compositeConfig.gsub!(/\%\{generator_list\}/, configString)

    prologString = ""
    if prologList.length > 0
      prologList.each do |savedProlog|
        prologString += "\n"
        prologString += savedProlog
      end
      prologString = "BEGIN_PROLOG" + prologString + "\nEND_PROLOG"
    end
    compositeConfig.gsub!(/\%\{prolog\}/, prologString)

    return compositeConfig
  end

  def generateXmlRpcClientList(cmdLineOptions)
    xmlrpcClients = ""
    (cmdLineOptions.v1495s).each do |proc|
      if proc.kind == "V1495"
        br = cmdLineOptions.boardReaders[proc.board_reader_index]
        if br.hasBeenIncludedInXMLRPCList
          next
        else
          br.hasBeenIncludedInXMLRPCList = true
          xmlrpcClients += ";http://" + proc.host + ":" +
            String(proc.port) + "/RPC2"
          xmlrpcClients += ",1"  # group number
        end
      end
    end
    (cmdLineOptions.v1495s).each do |proc|
      if proc.kind != "V1495"
        br = cmdLineOptions.boardReaders[proc.board_reader_index]
        if br.hasBeenIncludedInXMLRPCList
          next
        else
          br.hasBeenIncludedInXMLRPCList = true
          xmlrpcClients += ";http://" + proc.host + ":" +
            String(proc.port) + "/RPC2"
          xmlrpcClients += ",2"  # group number
        end
      end
    end
    (cmdLineOptions.v1720s + cmdLineOptions.v1190s).each do |proc|
      br = cmdLineOptions.boardReaders[proc.board_reader_index]
      if br.hasBeenIncludedInXMLRPCList
        next
      else
        br.hasBeenIncludedInXMLRPCList = true
        xmlrpcClients += ";http://" + proc.host + ":" +
          String(proc.port) + "/RPC2"
        xmlrpcClients += ",3"  # group number
      end
    end
    (cmdLineOptions.eventBuilders).each do |proc|
      xmlrpcClients += ";http://" + proc.host + ":" +
        String(proc.port) + "/RPC2"
      xmlrpcClients += ",4"  # group number
    end
    (cmdLineOptions.aggregators).each do |proc|
      xmlrpcClients += ";http://" + proc.host + ":" +
        String(proc.port) + "/RPC2"
      xmlrpcClients += ",5"  # group number
    end
    return xmlrpcClients
  end
end

class CommandLineParser
  def initialize(configGen)
    @configGen = configGen
    @options = OpenStruct.new
    @options.aggregators = []
    @options.eventBuilders = []
    @options.v1720s = []
    @options.v1190s = []
    @options.v1495s = []
    @options.boardReaders = []
    @options.dataDir = nil
    @options.command = nil
    @options.summary = false
    @options.runNumber = "0101"
    @options.serialize = false
    @options.runOnmon = 0
    @options.writeData = 1
    @options.runDurationSeconds = -1
    @options.eventsInRun = -1
    @options.fileSizeThreshold = 0
    @options.fileDurationSeconds = 0
    @options.eventsInFile = 0
    @options.vetoChoice = "off"
    @options.compressionTableFromDB = 0

    @optParser = OptionParser.new do |opts|
      opts.banner = "Usage: ds50MasterControl.rb [options]"
      opts.separator ""
      opts.separator "Specific options:"

      opts.on("--eb [host,port,compression_level]", Array,
              "Add an event builder that runs on the",
              "specified host and port and optionally",
              "compresses ADC data.",
              "For the compression level,",
              "  10 means do nothing at all",
              "  11 means compress and keep",
              "  12 means compress and drop",
              "  13 means do nothing and drop",
              "  14 means decompress") do |eb|
        if eb.length != 3
          puts "You must specifiy a host, port, and compression level."
          exit
        end
        ebConfig = OpenStruct.new
        ebConfig.host = eb[0]
        ebConfig.port = Integer(eb[1])
        ebConfig.compression_level = Integer(eb[2])
        ebConfig.kind = "eb"
        ebConfig.index = @options.eventBuilders.length
        @options.eventBuilders << ebConfig
      end

      opts.on("--ag [host,port,bunch_size,compression_level]", Array,
              "Add an aggregator that runs on the",
              "specified host and port.  Also specify the",
              "number of events to pass to art per bunch,",
              "and the compression level.",
              "For the compression level,",
              "  10 means do nothing at all",
              "  11 means compress and keep",
              "  12 means compress and drop",
              "  13 means do nothing and drop",
              "  14 means decompress") do |ag|
        if ag.length != 4
          puts "You must specifiy a host, port, bunch size, and"
          puts "compression level."
          exit
        end
        agConfig = OpenStruct.new
        agConfig.host = ag[0]
        agConfig.port = Integer(ag[1])
        agConfig.kind = "ag"
        agConfig.bunch_size = Integer(ag[2])
        agConfig.compression_level = Integer(ag[3])
        agConfig.index = @options.aggregators.length
        @options.aggregators << agConfig
      end
    
      opts.on("--v1495 [host,port,link_number,pulser_rate,laser_rate,random_trigger,gate_width,low_threshold,fnal_mode,board_id,g2_mode,inhibit_time,tdc_delay]", Array, 
              "Add a V1495 fragment receiver that runs on the specified host,",
              "port, and link, and has the specified parameters for pulser rate ",
              "laser rate, random trigger enable, gate width, majority logic ",
              "low threshold, FNAL mode selection flag, board ID, G2 mode,",
              "inhibit time, and TDC pulse delay.") do |v1495|
        if v1495.length != 13
          puts "You must specifiy a host, port, link_number, pulser rate,"
          puts "laser rate, random trigger enable, gate width majority logic,"
          puts "low threshold, FNAL mode flag, board ID, G2 mode, inhibit time,"
          puts "and TDC pulse delay for the V1495."
          exit
        end
        v1495Config = OpenStruct.new
        v1495Config.host = v1495[0]
        v1495Config.port = Integer(v1495[1])
        v1495Config.link_number = Integer(v1495[2])
        v1495Config.pulser_rate = Float(v1495[3])
        v1495Config.laser_rate = Float(v1495[4])
        v1495Config.random_trigger_enable = Integer(v1495[5])
        v1495Config.gate_width = Float(v1495[6])
        v1495Config.low_threshold = Integer(v1495[7])
        v1495Config.fnal_mode = Integer(v1495[8])
        v1495Config.board_id = Integer(v1495[9])
        v1495Config.g2_mode = Integer(v1495[10])
        v1495Config.inhibit_time = Float(v1495[11])
        v1495Config.tdc_pulse_delay = Float(v1495[12])
        v1495Config.kind = "V1495"
        v1495Config.index = @options.v1495s.length
        v1495Config.board_reader_index = addToBoardReaderList(v1495Config.host, v1495Config.port,
                                                              v1495Config.kind, v1495Config.index)
        @options.v1495s << v1495Config
      end

      opts.on("--v1495-veto [host,port,link_number,gate_width,tdc_pulse_delay,fanout_vme_base,board_id]", Array, 
              "Add a V1495Veto fragment receiver that runs on the specified host,",
              "port, and link, and has the specified parameters for gate width, ",
              "TDC pulse delay, fanout VME base address, and board ID.") do |v1495|
        if v1495.length != 7
          puts "You must specifiy a host, port, link_number, gate width,"
          puts "TDC pulse delay, fanout VME base and board ID for the V1495Veto."
          exit
        end
        v1495Config = OpenStruct.new
        v1495Config.host = v1495[0]
        v1495Config.port = Integer(v1495[1])
        v1495Config.link_number = Integer(v1495[2])
        v1495Config.gate_width = Float(v1495[3])
        v1495Config.tdc_pulse_delay = Float(v1495[4])
        v1495Config.fanout_vme_base = v1495[5]
        v1495Config.board_id = Integer(v1495[6])
        v1495Config.kind = "V1495-Veto"
        v1495Config.index = @options.v1495s.length
        v1495Config.board_reader_index = addToBoardReaderList(v1495Config.host, v1495Config.port,
                                                              v1495Config.kind, v1495Config.index)
        @options.v1495s << v1495Config
      end

      opts.on("--v1720 [host,port,link_number,sim,gate_width,pretrigger_window,board_id]", Array, 
              "Add a V1720 fragment receiver that runs on the specified host, port, and",
              "link, and has the specified gate width, pretrigger window, and board ID.",
              "The sim parameter (0 or 1) determines whether or not to use the",
              "V1720 simulator.") do |v1720|
        if v1720.length != 7
          puts "You must specifiy a host, port, and link_number, whether or not the"
          puts "V1720 is simulated, the gate width and pre-trigger window (usec),"
          puts "and the board ID."
          exit
        end
        v1720Config = OpenStruct.new
        v1720Config.host = v1720[0]
        v1720Config.port = Integer(v1720[1])
        v1720Config.link_number = Integer(v1720[2])
        v1720Config.sim = Integer(v1720[3])
        v1720Config.gate_width = Integer(v1720[4])
        v1720Config.pretrigger_window = Float(v1720[5])
        v1720Config.board_id = Integer(v1720[6])
        v1720Config.kind = "V1720"
        v1720Config.index = @options.v1720s.length
        v1720Config.board_reader_index = addToBoardReaderList(v1720Config.host, v1720Config.port,
                                                              v1720Config.kind, v1720Config.index)
        @options.v1720s << v1720Config
      end

      opts.on("--v1724 [host,port,link_number,sim,gate_width,pretrigger_window,board_id]", Array, 
              "Add a V1724 fragment receiver that runs on the specified host, port, and",
              "link, and has the specified gate width, pretrigger window, and board ID.",
              "The sim parameter (0 or 1) determines whether or",
              "not to use the V1724 simulator.") do |v1724|
        if v1724.length != 7
          puts "You must specifiy a host, port, link_number, whether or not the"
          puts "V1724 is simulated, the gate width and pre-trigger window (usec),"
          puts "and the board ID."
          exit
        end
        v1724Config = OpenStruct.new
        v1724Config.host = v1724[0]
        v1724Config.port = Integer(v1724[1])
        v1724Config.link_number = Integer(v1724[2])
        v1724Config.sim = Integer(v1724[3])
        v1724Config.gate_width = Integer(v1724[4])
        v1724Config.pretrigger_window = Float(v1724[5])
        v1724Config.board_id = Integer(v1724[6])
        v1724Config.kind = "V1724"
        v1724Config.index = @options.v1720s.length
        v1724Config.board_reader_index = addToBoardReaderList(v1724Config.host, v1724Config.port,
                                                              v1724Config.kind, v1724Config.index)
        # NOTE that we're simply adding this to the 1720 list...
        @options.v1720s << v1724Config
      end

      opts.on("--v1190 [host,port,link_number,vme_base,gate_width,post_trigger,irq_level,irq_vector,board_id]", Array, 
              "Add a V1190 fragment receiver that runs on the specified host, ",
              "port and link with the specified VME base address, gate width, ",
              "post-trigger window, irq_level, irq_vector, and board ID.") do |v1190|
        if v1190.length != 9
          puts "You must specifiy a host, port, link_number, vme_base, gate_width,"
          puts "post_trigger, irq_level, irq_vector, and board ID."
          exit
        end
        v1190Config = OpenStruct.new
        v1190Config.host = v1190[0]
        v1190Config.port = Integer(v1190[1])
        v1190Config.link_number = Integer(v1190[2])
        v1190Config.vme_base = v1190[3]
        v1190Config.gate_width = Float(v1190[4])
        v1190Config.post_trigger = Float(v1190[5])
        v1190Config.irq_level = Integer(v1190[6])
        v1190Config.irq_vector = Integer(v1190[7])
        v1190Config.board_id = Integer(v1190[8])
        v1190Config.kind = "V1190"
        v1190Config.index = @options.v1190s.length
        v1190Config.board_reader_index = addToBoardReaderList(v1190Config.host, v1190Config.port,
                                                              v1190Config.kind, v1190Config.index)
        @options.v1190s << v1190Config
      end

      opts.on("-d", "--data-dir [data dir]", 
              "Directory that the event builders will", "write data to.") do |dataDir|
        @options.dataDir = dataDir
      end

      opts.on("-m", "--online-monitoring [enable flag (0 or 1)]", 
              "Whether to run the online monitoring modules.") do |runOnmon|
        @options.runOnmon = runOnmon
      end

      opts.on("-w", "--write-data [enable flag (0 or 1)]", 
              "Whether to write data to disk.") do |writeData|
        @options.writeData = writeData
      end

      opts.on("-c", "--command [command]", 
              "Execute a command: start, stop, init, shutdown, pause, resume, status, get-legal-commands.") do |command|
        @options.command = command
      end

      opts.on("-r", "--run-number [number]", "Specify the run number.") do |run|
        @options.runNumber = run
      end

      opts.on("-t", "--run-duration [minutes]",
              "Stop the run after the specified amount of time (minutes).") do |timeInMinutes|
        @options.runDurationSeconds = Integer(timeInMinutes) * 60
      end

      opts.on("-n", "--run-event-count [number]",
              "Stop the run after the specified number of events have been collected.") do |eventCount|
        @options.eventsInRun = Integer(eventCount)
      end

      opts.on("-f", "--file-size [number of MB]",
              "Close each data file when the specified size is reached (MB).") do |fileSize|
        @options.fileSizeThreshold = Float(fileSize)
      end

      opts.on("--file-duration [minutes]",
              "Closes each file after the specified amount of time (minutes).") do |timeInMinutes|
        @options.fileDurationSeconds = Integer(timeInMinutes) * 60
      end

      opts.on("--file-event-count [number]",
              "Close each file after the specified number of events have been written.") do |eventCount|
        @options.eventsInFile = Integer(eventCount)
      end

      opts.on("-s", "--summary", "Summarize the configuration.") do
        @options.summary = true
      end

      opts.on("-e", "--serialize", "Serialize the config for System Control.") do
        @options.serialize = true
      end

      opts.on("--veto-choice [choice]",
              "Veto configuration choice (\"off\", \"TDC\", \"ADC\", \"Hybrid\", \"TDC+ADC\", \"TDC+Hybrid\").") do |choice|
        @options.vetoChoice = choice
      end

      opts.on("--compression-table-from-db [enable flag (0 or 1)]",
              "Whether to fetch the compression tables from the database.") do |dbTable|
        @options.compressionTableFromDB = dbTable
      end

      opts.on_tail("-h", "--help", "Show this message.") do
        puts opts
        exit
      end
    end
  end

  def parse()
    if ARGV.length == 0
      puts @optParser
      exit
    end

    @optParser.parse(ARGV)
    self.validate()
  end

  def validate()
    # Make sure link numbers are unique
    # In data mode make sure there is an eb, 1720, 1190 and 1495
    # In sim mode make sure there is an eb and a 1720
    return nil
  end

  def addToBoardReaderList(host, port, kind, boardIndex)
    # check for an existing boardReader with the same host and port
    brIndex = 0
    @options.boardReaders.each do |br|
      if host == br.host && port == br.port
        br.kindList << kind
        br.boardIndexList << boardIndex
        br.cfgList << ""
        br.boardCount += 1
        return brIndex
      end
      brIndex += 1
    end

    # if needed, create a new boardReader
    br = OpenStruct.new
    br.host = host
    br.port = port
    br.kindList = [kind]
    br.boardIndexList = [boardIndex]
    br.cfgList = [""]
    br.boardCount = 1
    br.commandHasBeenSent = false
    br.hasBeenIncludedInXMLRPCList = false
    br.kind = "multi-board"

    brIndex = @options.boardReaders.length
    @options.boardReaders << br
    return brIndex
  end

  def summarize()
    # Print out a summary of the configuration that was passed in from the
    # the command line.  Everything will be printed in terms of what process
    # is running on which host.
    puts "Configuration Summary:"
    hostMap = {}
    (@options.eventBuilders + @options.v1495s + @options.v1720s + 
     @options.aggregators + @options.v1190s).each do |proc|
      if not hostMap.keys.include?(proc.host)
        hostMap[proc.host] = []
      end
      hostMap[proc.host] << proc
    end

    hostMap.each_key { |host|
      puts "  %s:" % [host]
      hostMap[host].sort! { |x,y|
        x.port <=> y.port
      }

      totalEBs = @options.eventBuilders.length
      #totalFRs = @options.v1495s.length + @options.v1720s.length + @options.v1190s.length
      totalFRs = @options.boardReaders.length
      hostMap[host].each do |item|
        case item.kind
        when "eb"
          puts "    EventBuilder, port %d, rank %d" % [item.port, totalFRs + item.index]
        when "ag"
          puts "    Aggregator, port %d, rank %d" % [item.port, totalEBs + totalFRs + item.index]
        when "V1495","V1495-Veto"
          puts "    BoardReader, V1495, port %s, link %s, rank %d" % [item.port, 
                                                                           item.link_number,
                                                                           item.index]
        when "V1720", "V1724"
          if item.sim != 0
            puts "    BoardReader, Simulated %s, port %d, rank %d" % [item.kind.upcase,
                                                                           item.port,
                                                                           @options.v1495s.length +
                                                                           item.index]
          else
            puts "    BoardReader, %s, port %d, link %d, rank %d" % [item.kind.upcase,
                                                                          item.port, 
                                                                          item.link_number,
                                                                          @options.v1495s.length +
                                                                          item.index]
          end
        when "V1190"
          puts "    BoardReader, V1190, port %s, link %s, rank %d" % [item.port, 
                                                                           item.link_number,
                                                                           @options.v1495s.length +
                                                                           @options.v1720s.length +
                                                                           item.index]
        end
      end
      puts ""
    }
    STDOUT.flush
    return nil
  end

  def getOptions()
    return @options
  end
end

class SystemControl
  def initialize(cmdLineOptions, configGen)
    @options = cmdLineOptions
    @configGen = configGen
  end

  def init()
    ebIndex = 0
    agIndex = 0
    totalv1720s = 0
    totalv1724s = 0
    @options.v1720s.each do |proc|
      case proc.kind
      when "V1720"
        totalv1720s += 1
      when "V1724"
        totalv1724s += 1
      end
    end
    totalBoards = @options.v1495s.length + @options.v1720s.length + @options.v1190s.length
    totalFRs = @options.boardReaders.length
    totalEBs = @options.eventBuilders.length
    totalAGs = @options.aggregators.length
    inputBuffSizeWords = 2097152
    if @options.v1495s.length > 0
      inputBuffSizeWords = 8192 * @options.v1495s[0].gate_width
    elsif Integer(totalv1720s) > 0
      inputBuffSizeWords = 8192 * @options.v1720s[0].gate_width
    end
    xmlrpcClients = @configGen.generateXmlRpcClientList(@options)

    # 02-Dec-2013, KAB - loop over the front-end boards and build the configurations
    # that we will send to them.  These configurations are stored in the associated
    # boardReaders list entries since there are system configurations in which
    # multiple boards are read out by a single BoardReader, and it seems simpler to
    # store the CFGs in the boardReader list for everything
    @options.v1495s.each { |v1495Options|
      br = @options.boardReaders[v1495Options.board_reader_index]
      listIndex = 0
      br.kindList.each do |kind|
        if kind == v1495Options.kind && br.boardIndexList[listIndex] == v1495Options.index
          if v1495Options.kind == "V1495-Veto"
            cfg = @configGen.generateV1495Veto(v1495Options.index,
                                               totalEBs, totalFRs, 
                                               v1495Options.link_number,
                                               v1495Options.gate_width,
                                               v1495Options.tdc_pulse_delay,
                                               v1495Options.fanout_vme_base,
                                               @options.vetoChoice,
                                               Integer(inputBuffSizeWords/8),
                                               v1495Options.board_id)
          else
            cfg = @configGen.generateV1495(v1495Options.index,
                                           totalEBs, totalFRs, 
                                           v1495Options.link_number,
                                           v1495Options.pulser_rate,
                                           v1495Options.laser_rate,
                                           v1495Options.random_trigger_enable,
                                           v1495Options.gate_width,
                                           v1495Options.low_threshold,
                                           v1495Options.fnal_mode,
                                           @options.vetoChoice,
                                           Integer(inputBuffSizeWords/8),
                                           v1495Options.board_id,
                                           v1495Options.g2_mode,
                                           v1495Options.inhibit_time,
                                           v1495Options.tdc_pulse_delay)
          end
          br.cfgList[listIndex] = cfg
          break
        end
        listIndex += 1
      end
    }
    @options.v1720s.each { |v1720Options|
      br = @options.boardReaders[v1720Options.board_reader_index]
      listIndex = 0
      br.kindList.each do |kind|
        if kind == v1720Options.kind && br.boardIndexList[listIndex] == v1720Options.index
          cfg = @configGen.generateV1720(@options.v1495s.length + v1720Options.index,
                                         totalEBs, totalFRs, v1720Options.link_number,
                                         v1720Options.sim, v1720Options.gate_width,
                                         v1720Options.pretrigger_window,
                                         Integer(inputBuffSizeWords/8),
                                         v1720Options.board_id)
          br.cfgList[listIndex] = cfg
          break
        end
        listIndex += 1
      end
    }
    @options.v1190s.each { |v1190Options|
      br = @options.boardReaders[v1190Options.board_reader_index]
      listIndex = 0
      br.kindList.each do |kind|
        if kind == v1190Options.kind && br.boardIndexList[listIndex] == v1190Options.index
          cfg = @configGen.generateV1190(@options.v1495s.length +
                                         @options.v1720s.length + v1190Options.index,
                                         totalEBs, totalFRs, v1190Options.link_number,
                                         v1190Options.vme_base, v1190Options.gate_width,
                                         v1190Options.post_trigger, v1190Options.irq_level,
                                         v1190Options.irq_vector,
                                         Integer(inputBuffSizeWords/8),
                                         v1190Options.board_id)
          br.cfgList[listIndex] = cfg
          break
        end
        listIndex += 1
      end
    }

    ## debug
    #@options.boardReaders.each { |br|
    #  puts "=========="
    #  puts br.host
    #  puts br.port
    #  puts br.boardCount
    #  listIndex = 0
    #  br.kindList.each do |kind|
    #    puts " *** "
    #    puts kind
    #    puts br.boardIndexList[listIndex]
    #    puts br.cfgList[listIndex]
    #    listIndex += 1
    #  end
    #}

    # 27-Jun-2013, KAB - send INIT to 1495s first
    threads = []
    @options.v1495s.each { |proc|
      br = @options.boardReaders[proc.board_reader_index]
      if br.boardCount > 1
        if br.commandHasBeenSent
          next
        else
          br.commandHasBeenSent = true
          proc = br
          cfg = @configGen.generateComposite(totalEBs, totalFRs, br.cfgList)
        end
      else
        cfg = br.cfgList[0]
      end

      currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
      puts "%s: Sending the INIT command to %s:%d." %
        [currentTime, proc.host, proc.port]
      threads << Thread.new() do
        xmlrpcClient = XMLRPC::Client.new(proc.host, "/RPC2", 
                                          proc.port)
        if @options.serialize
          fileName = "BoardReader_%s_%s_%d.fcl" % [proc.kind, proc.host, proc.port]
          puts "  writing %s..." % fileName
          handle = File.open(fileName, "w")
          handle.write(cfg.gsub(/\n/,"\n\r"))
          handle.write("\n\r");
          handle.close()
        end
        result = xmlrpcClient.call("daq.init", cfg)
        currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
        puts "%s: %s BoardReader on %s:%d result: %s" %
          [currentTime, proc.kind, proc.host, proc.port, result]
        STDOUT.flush
      end
    }
    STDOUT.flush
    threads.each { |aThread|
      aThread.join()
    }

    # 27-Jun-2013, KAB - send INIT to other hardware next
    threads = []
    @options.v1720s.each { |proc|
      br = @options.boardReaders[proc.board_reader_index]
      if br.boardCount > 1
        if br.commandHasBeenSent
          next
        else
          br.commandHasBeenSent = true
          proc = br
          cfg = @configGen.generateComposite(totalEBs, totalFRs, br.cfgList)
        end
      else
        cfg = br.cfgList[0]
      end

      currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
      puts "%s: Sending the INIT command to %s:%d." %
        [currentTime, proc.host, proc.port]
      threads << Thread.new() do
        xmlrpcClient = XMLRPC::Client.new(proc.host, "/RPC2",
                                          proc.port)
        if @options.serialize
          fileName = "BoardReader_%s_%s_%d.fcl" % [proc.kind,proc.host, proc.port]
          puts "  writing %s..." % fileName
          handle = File.open(fileName, "w")
          handle.write(cfg.gsub(/\n/,"\n\r"))
          handle.write("\n\r");
          handle.close()
        end
        result = xmlrpcClient.call("daq.init", cfg)
        currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
        puts "%s: %s BoardReader on %s:%d result: %s" %
          [currentTime, proc.kind, proc.host, proc.port, result]
        STDOUT.flush
      end
    }
    @options.v1190s.each { |proc|
      br = @options.boardReaders[proc.board_reader_index]
      if br.boardCount > 1
        if br.commandHasBeenSent
          next
        else
          br.commandHasBeenSent = true
          proc = br
          cfg = @configGen.generateComposite(totalEBs, totalFRs, br.cfgList)
        end
      else
        cfg = br.cfgList[0]
      end

      currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
      puts "%s: Sending the INIT command to %s:%d." %
        [currentTime, proc.host, proc.port]
      threads << Thread.new() do
        xmlrpcClient = XMLRPC::Client.new(proc.host, "/RPC2",
                                          proc.port)
        if @options.serialize
          fileName = "BoardReader_%s_%s_%d.fcl" % [proc.kind, proc.host, proc.port]
          puts "  writing %s..." % fileName
          handle = File.open(fileName, "w")
          handle.write(cfg.gsub(/\n/,"\n\r"))
          handle.write("\n\r");
          handle.close()
        end
        result = xmlrpcClient.call("daq.init", cfg)
        currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
        puts "%s: %s BoardReader on %s:%d result: %s" %
          [currentTime, proc.kind, proc.host, proc.port, result]
        STDOUT.flush
      end
    }
    STDOUT.flush
    threads.each { |aThread|
      aThread.join()
    }

    # 27-Jun-2013, KAB - send INIT to EBs and AG last
    threads = []
    @options.eventBuilders.each { |ebOptions|
      currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
      puts "%s: Sending the INIT command to %s:%d." %
        [currentTime, ebOptions.host, ebOptions.port]
      threads << Thread.new() do
        xmlrpcClient = XMLRPC::Client.new(ebOptions.host, "/RPC2", 
                                          ebOptions.port)
        cfg = @configGen.generateEventBuilder(ebIndex, totalFRs, totalEBs, totalAGs,
                                              ebOptions.compression_level,
                                              totalv1720s, totalv1724s,
                                              @options.dataDir, @options.runOnmon,
                                              @options.writeData, inputBuffSizeWords,
                                              totalBoards, @options.compressionTableFromDB)
        if @options.serialize
          fileName = "EventBuilder_%s_%d.fcl" % [ebOptions.host, ebOptions.port]
          puts "  writing %s..." % fileName
          handle = File.open(fileName, "w")
          handle.write(cfg.gsub(/\n/,"\n\r"))
          handle.write("\n\r");
          handle.close()
        end
        result = xmlrpcClient.call("daq.init", cfg)
        currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
        puts "%s: EventBuilder on %s:%d result: %s" %
          [currentTime, ebOptions.host, ebOptions.port, result]
        STDOUT.flush
      end
      ebIndex += 1
    }

    @options.aggregators.each { |agOptions|
      currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
      puts "%s: Sending the INIT command to %s:%d." %
        [currentTime, agOptions.host, agOptions.port]
      threads << Thread.new() do
        xmlrpcClient = XMLRPC::Client.new(agOptions.host, "/RPC2", 
                                          agOptions.port)
        cfg = @configGen.generateAggregator(@options.dataDir, @options.runNumber,
                                            totalFRs, totalEBs, agOptions.bunch_size,
                                            agOptions.compression_level,
                                            totalv1720s, totalv1724s,
                                            @options.runOnmon, @options.writeData,
                                            agIndex, totalAGs, inputBuffSizeWords,
                                            xmlrpcClients, @options.fileSizeThreshold,
                                            @options.fileDurationSeconds,
                                            @options.eventsInFile,
                                            @options.compressionTableFromDB)
        if @options.serialize
          fileName = "Aggregator_%s_%d.fcl" % [agOptions.host, agOptions.port]
          puts "  writing %s..." % fileName
          handle = File.open(fileName, "w")
          handle.write(cfg.gsub(/\n/,"\n\r"))
          handle.write("\n\r");
          handle.close()
        end
        result = xmlrpcClient.call("daq.init", cfg)
        currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
        puts "%s: Aggregator on %s:%d result: %s" %
          [currentTime, agOptions.host, agOptions.port, result]
        STDOUT.flush
      end
      agIndex += 1
    }
    STDOUT.flush
    threads.each { |aThread|
      aThread.join()
    }
  end

  def start(runNumber)
    self.sendCommandSet("start", @options.aggregators, runNumber)
    self.sendCommandSet("start", @options.eventBuilders, runNumber)
    self.sendCommandSet("start", @options.v1720s + @options.v1190s, runNumber)
    v1495List = []
    v1495VetoList = []
    @options.v1495s.each do |proc|
      if proc.kind == "V1495-Veto"
        v1495VetoList << proc
      else
        v1495List << proc
      end
    end
    self.sendCommandSet("start", v1495VetoList, runNumber)
    self.sendCommandSet("start", v1495List, runNumber)
  end

  def sendCommandSet(commandName, procs, commandArg = nil)
    threads = []
    procs.each do |proc|
      # 02-Dec-2013, KAB - use the boardReader instance instead of the
      # actual card when multiple cards are read out by a single BR
      if proc.board_reader_index != nil
        br = @options.boardReaders[proc.board_reader_index]
        if br.boardCount > 1
          if br.commandHasBeenSent
            next
          else
            br.commandHasBeenSent = true
            proc = br
          end
        end
      end

      currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
      puts "%s: Attempting to connect to %s:%d and %s a run." %
        [currentTime, proc.host, proc.port, commandName]
      STDOUT.flush
      threads << Thread.new() do
        xmlrpcClient = XMLRPC::Client.new(proc.host, "/RPC2", proc.port)
        xmlrpcClient.timeout = 60
        if commandName == "stop"
          if proc.kind == "ag"
            xmlrpcClient.timeout = 120
          elsif proc.kind == "eb" || proc.kind == "multi-board"
            xmlrpcClient.timeout = 45
          else
            xmlrpcClient.timeout = 30
          end
        end
        begin
          if commandArg != nil
            result = xmlrpcClient.call("daq.%s" % [commandName], commandArg)
          else
            result = xmlrpcClient.call("daq.%s" % [commandName])
          end
        rescue Exception => msg
          result = "Exception: " + msg
        end
        currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
        case proc.kind
        when "eb"
          puts "%s: EventBuilder on %s:%d result: %s" %
            [currentTime, proc.host, proc.port, result]
        when "ag"
          puts "%s: Aggregator on %s:%d result: %s" %
            [currentTime, proc.host, proc.port, result]
        when "V1495", "V1495-Veto"
          puts "%s: V1495 BoardReader on %s:%d result: %s" %
            [currentTime, proc.host, proc.port, result]
        when "V1720"
          puts "%s: V1720 BoardReader on %s:%d result: %s" %
            [currentTime, proc.host, proc.port, result]
        when "V1724"
          puts "%s: V1724 BoardReader on %s:%d result: %s" %
            [currentTime, proc.host, proc.port, result]
        when "V1190"
          puts "%s: V1190 BoardReader on %s:%d result: %s" %
            [currentTime, proc.host, proc.port, result]
        when "multi-board"
          puts "%s: multi-board BoardReader on %s:%d result: %s" %
            [currentTime, proc.host, proc.port, result]
        end
        STDOUT.flush
      end
    end
    threads.each { |aThread|
      aThread.join()
    }
  end

  def shutdown()
    v1495List = []
    v1495VetoList = []
    @options.v1495s.each do |proc|
      if proc.kind == "V1495-Veto"
        v1495VetoList << proc
      else
        v1495List << proc
      end
    end
    self.sendCommandSet("shutdown", v1495List)
    self.sendCommandSet("shutdown", v1495VetoList)
    self.sendCommandSet("shutdown", @options.v1720s + @options.v1190s)
    self.sendCommandSet("shutdown", @options.eventBuilders)
    self.sendCommandSet("shutdown", @options.aggregators)
  end

  def pause()
    v1495List = []
    v1495VetoList = []
    @options.v1495s.each do |proc|
      if proc.kind == "V1495-Veto"
        v1495VetoList << proc
      else
        v1495List << proc
      end
    end
    self.sendCommandSet("pause", v1495List)
    self.sendCommandSet("pause", v1495VetoList)
    self.sendCommandSet("pause", @options.v1720s + @options.v1190s)
    self.sendCommandSet("pause", @options.eventBuilders)
    self.sendCommandSet("pause", @options.aggregators)
  end

  def stop()
    totalAGs = @options.aggregators.length
    if @options.eventsInRun > 0
      if Integer(totalAGs) > 0
        if Integer(totalAGs) > 1
          puts "NOTE: more than one Aggregator is running (count=%d)." % [totalAGs]
          puts " -> The first Aggregator will be used to determine the number of events"
          puts " -> in the current run."
        end
        aggregatorEventCount = 0
        previousAGEventCount = 0
        sleepTime = 0
        while aggregatorEventCount >= 0 && aggregatorEventCount < @options.eventsInRun do
          sleep(sleepTime)
          currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
          puts "%s: Attempting to fetch the number of events from the Aggregator." %
            [currentTime]
          STDOUT.flush
          xmlrpcClient = XMLRPC::Client.new(@options.aggregators[0].host, "/RPC2",
                                            @options.aggregators[0].port)
          xmlrpcClient.timeout = 10
          exceptionOccurred = false
          begin
            result = xmlrpcClient.call("daq.report", "event_count")
            if result == "busy" || result == "-1"
              # support one retry
              sleep(10)
              result = xmlrpcClient.call("daq.report", "event_count")
            end
            aggregatorEventCount = Integer(result)
          rescue Exception => msg
            exceptionOccurred = true
            result = "Exception: " + msg
            aggregatorEventCount = previousAGEventCount
          end
          currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
          if exceptionOccurred
            puts "%s: There was a problem communicating with the Aggregator (%s)," %
              [currentTime, result]
            puts "  the fetch of the number of events will be retried."
          else
            puts "%s: The Aggregator reports the following number of events: %s." %
              [currentTime, result]
          end
          STDOUT.flush

          if aggregatorEventCount > 0 && previousAGEventCount > 0 && \
            aggregatorEventCount > previousAGEventCount && sleepTime > 0 then
            remainingEvents = @options.eventsInRun - aggregatorEventCount
            recentRate = (aggregatorEventCount - previousAGEventCount) / sleepTime
            if recentRate > 0
              sleepTime = (remainingEvents / 2) / recentRate
            else
              sleepTime = 10
            end
            if sleepTime < 10
              sleepTime = 10;
            end
            if sleepTime > 900
              sleepTime = 900;
            end
          else
            sleepTime = 10
          end
          previousAGEventCount = aggregatorEventCount
        end
      else
        puts "No Aggregator in use - Unable to determine the number of events in the current run."     
      end
    elsif @options.runDurationSeconds > 0
      if Integer(totalAGs) > 0
        if Integer(totalAGs) > 1
          puts "NOTE: more than one Aggregator is running (count=%d)." % [totalAGs]
          puts " -> The first Aggregator will be used to determine the run duration."
        end
        aggregatorRunDuration = 0
        sleepTime = 0
        while aggregatorRunDuration >= 0 && aggregatorRunDuration < @options.runDurationSeconds do
          sleep(sleepTime)
          currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
          puts "%s: Attempting to fetch the run duration from the Aggregator." %
            [currentTime]
          STDOUT.flush
          xmlrpcClient = XMLRPC::Client.new(@options.aggregators[0].host, "/RPC2",
                                            @options.aggregators[0].port)
          xmlrpcClient.timeout = 10
          exceptionOccurred = false
          begin
            result = xmlrpcClient.call("daq.report", "run_duration")
            if result == "busy" || result == "-1"
              # support one retry
              sleep(10)
              result = xmlrpcClient.call("daq.report", "run_duration")
            end
            aggregatorRunDuration = Float(result)
          rescue Exception => msg
            exceptionOccurred = true
            result = "Exception: " + msg
            aggregatorRunDuration = 0
          end
          currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
          if exceptionOccurred
            puts "%s: There was a problem communicating with the Aggregator (%s)," %
              [currentTime, result]
            puts "  the fetch of the run duration will be retried."
          else
            puts "%s: The Aggregator reports the following run duration: %s seconds." %
              [currentTime, result]
          end
          STDOUT.flush

          if aggregatorRunDuration > 0 then
            remainingTime = @options.runDurationSeconds - aggregatorRunDuration
            sleepTime = remainingTime / 2
            if sleepTime < 10
              sleepTime = 10;
            end
            if sleepTime > 900
              sleepTime = 900;
            end
          else
            sleepTime = 10
          end
        end
      else
        puts "No Aggregator in use - Unable to determine the duration of the current run."
      end
    end

    v1495List = []
    v1495VetoList = []
    @options.v1495s.each do |proc|
      if proc.kind == "V1495-Veto"
        v1495VetoList << proc
      else
        v1495List << proc
      end
    end
    self.sendCommandSet("stop", v1495List)
    self.sendCommandSet("stop", v1495VetoList)
    self.sendCommandSet("stop", @options.v1720s + @options.v1190s)
    self.sendCommandSet("stop", @options.eventBuilders)
    @options.aggregators.each do |proc|
      tmpList = []
      tmpList << proc
      self.sendCommandSet("stop", tmpList)
    end
  end

  def resume()
    self.sendCommandSet("resume", @options.aggregators)
    self.sendCommandSet("resume", @options.eventBuilders)
    self.sendCommandSet("resume", @options.v1720s + @options.v1190s)
    v1495List = []
    v1495VetoList = []
    @options.v1495s.each do |proc|
      if proc.kind == "V1495-Veto"
        v1495VetoList << proc
      else
        v1495List << proc
      end
    end
    self.sendCommandSet("resume", v1495VetoList)
    self.sendCommandSet("resume", v1495List)
  end

  def checkStatus()
    self.sendCommandSet("status", @options.aggregators)
    self.sendCommandSet("status", @options.eventBuilders)
    self.sendCommandSet("status", @options.v1720s + @options.v1190s)
    self.sendCommandSet("status", @options.v1495s)
  end

  def getLegalCommands()
    self.sendCommandSet("legal_commands", @options.aggregators)
    self.sendCommandSet("legal_commands", @options.eventBuilders)
    self.sendCommandSet("legal_commands", @options.v1720s + @options.v1190s)
    self.sendCommandSet("legal_commands", @options.v1495s)
  end
end

if __FILE__ == $0
  cfgGen = ConfigGen.new
  cmdLineParser = CommandLineParser.new(cfgGen)
  cmdLineParser.parse()
  options = cmdLineParser.getOptions()
  puts "ds50MasterControl disk writing setting = " + options.writeData

  if options.summary
    cmdLineParser.summarize()
  end

  sysCtrl = SystemControl.new(options, cfgGen)

  if options.command == "init"
    sysCtrl.init()
  elsif options.command == "start"
    sysCtrl.start(options.runNumber)
  elsif options.command == "stop"
    sysCtrl.stop()
  elsif options.command == "shutdown"
    sysCtrl.shutdown()
  elsif options.command == "pause"
    sysCtrl.pause()
  elsif options.command == "resume"
    sysCtrl.resume()
  elsif options.command == "status"
    sysCtrl.checkStatus()
  elsif options.command == "get-legal-commands"
    sysCtrl.getLegalCommands()
  end
end
