#!/bin/bash

source `which setPortNumbers.sh`

AGGREGATOR_NODE="dsag"
THIS_NODE=`hostname -s`

# this function expects a number of arguments:
#  1) the DAQ command to be sent
#  2) the run number (dummy for non-start commands)
#  3) the pulser trigger rate
#  4) the laser trigger rate
#  5) whether to enable the random trigger (0 or 1)
#  6) the acqusition gate width (usec)
#  7) the majority logic low threshold
#  8) the compression level for ADC data [0..2]
#  9) whether to run online monitoring
# 10) the data directory
# 11) the logfile name
# 12) whether to write data to disk [0,1]
# 13) the desired number of events in the run (stop command)
# 14) the desired duration of the run (minutes, stop command)
# 15) the desired size of each data file
# 16) the pre-trigger window size (usec)
# 17) the desired number of events in each file
# 18) the desired time duration of each file (minutes)
# 19) whether to print out CFG information (verbose)
# 20) the choice of which Veto components to include
# 21) whether to use the G2 trigger
# 22) whether to include the V1724 ADCs in the system
# 23) the inhibit time (usec)
# 24) the delay in the TDC stop pulse from the V1495 (usec)
# 25) the width of the TDC gate (usec)
function launch() {
  let ebComp=$8+10
  let dwAgComp=$8+10
  let onmonAgComp=$8+10
  if [[ "$8" == "3" ]]; then
    ebComp=12
    dwAgComp=10
    onmonAgComp=14
  elif [[ "$8" == "2" ]]; then
    ebComp=11
    dwAgComp=13
    onmonAgComp=10
  elif [[ "$8" == "1" ]]; then
    ebComp=11
    dwAgComp=10
    onmonAgComp=10
  fi
  enableSerial=""
  if [[ "${19}" == "1" ]]; then
      enableSerial="-e"
  fi

  vetoChoice="${20}"
  vetoCfg=""
  if [[ "$vetoChoice" != "off" ]]; then
      if [[ "`echo $vetoChoice | grep TDC`" != "" ]]; then
          vetoCfg="${vetoCfg} --v1190 dsfr4,$frPort0,0,0x81000000,${25},1,0,0,0"
          vetoCfg="${vetoCfg} --v1190 dsfr4,$frPort0,0,0x82000000,${25},1,0,0,1"
      fi
      vetoCfg="${vetoCfg} --v1495-veto dsfr4,$frPort0,0,$6,${24},0x02000000,1"
  fi
  v1724Enable="${22}"
  v1724Cfg=""
  if [[ "$v1724Enable" == "yes" ]]; then
      v1724Cfg="${v1724Cfg} --v1724 dsfr2,$frPort2,2,0,$6,${16},0"
      v1724Cfg="${v1724Cfg} --v1724 dsfr2,$frPort3,3,0,$6,${16},1"
      v1724Cfg="${v1724Cfg} --v1724 dsfr3,$frPort0,0,0,$6,${16},2"
      v1724Cfg="${v1724Cfg} --v1724 dsfr3,$frPort1,1,0,$6,${16},3"
      v1724Cfg="${v1724Cfg} --v1724 dsfr3,$frPort2,2,0,$6,${16},4"
  fi

  ds50MasterControl.rb ${enableSerial} -s -c $1 \
    --v1495 dsfr1,$frPort0,0,$3,$4,$5,$6,$7,0,0,${21},${23},${24} \
    --v1720 dsfr1,$frPort1,1,0,$6,${16},0 \
    --v1720 dsfr1,$frPort2,2,0,$6,${16},1 \
    --v1720 dsfr1,$frPort3,3,0,$6,${16},2 \
    --v1720 dsfr2,$frPort0,0,0,$6,${16},3 \
    --v1720 dsfr2,$frPort1,1,0,$6,${16},4 \
    ${v1724Cfg} \
    ${vetoCfg} \
    --eb dseb1,$ebPort0,$ebComp \
    --eb dseb1,$ebPort1,$ebComp \
    --eb dseb1,$ebPort2,$ebComp \
    --eb dseb1,$ebPort3,$ebComp \
    --eb dseb2,$ebPort0,$ebComp \
    --eb dseb2,$ebPort1,$ebComp \
    --eb dseb2,$ebPort2,$ebComp \
    --eb dseb2,$ebPort3,$ebComp \
    --eb dseb3,$ebPort0,$ebComp \
    --eb dseb3,$ebPort1,$ebComp \
    --eb dseb3,$ebPort2,$ebComp \
    --eb dseb3,$ebPort3,$ebComp \
    --eb dseb4,$ebPort0,$ebComp \
    --eb dseb4,$ebPort1,$ebComp \
    --eb dseb4,$ebPort2,$ebComp \
    --eb dseb4,$ebPort3,$ebComp \
    --ag ${AGGREGATOR_NODE},$agPort0,1,$dwAgComp \
    --ag ${AGGREGATOR_NODE},$agPort1,1,$onmonAgComp \
    --data-dir ${10} --online-monitoring $9 \
    --write-data ${12} --run-event-count ${13} \
    --run-duration ${14} --file-size ${15} \
    --file-event-count ${17} --file-duration ${18} \
    --veto-choice ${20} \
    --compression-table-from-db 0 \
    --run-number $2 2>&1 | tee -a ${11}
}

scriptName=`basename $0`
usage () {
    echo "
Usage: ${scriptName} [options] <command> [command options]
Where command is one of:
  init, start, pause, resume, stop, status, get-legal-commands,
  shutdown, start-system, restart, reinit, exit,
  fast-shutdown, fast-restart, fast-reinit, or fast-exit
General options:
  -h, --help: prints this usage message
Configuration options (init commands):
  -t <low threshold>: specifies the low threshold for the majority logic
      trigger [default=5; zero means that majority logic trigger is off]
  -l <laser rate>: specifies the rate of the laser trigger
      [default=0, which means that the laser trigger is off]
  -p <pulser rate>: specifies the rate of the pulser trigger
      [default=0, which means that the pulser trigger is off]
  -r : enables the random trigger [default is random trigger off]
  -g <gate width>: specifies the acquisition gate width in usec [default=400]
  -i <inhibit time>: specifies the inhibit time in usec
      [default is a function of the gate width]
  -w <pre-trigger window>: specifies the size of the pre-trigger window in usec
      [default is 1 usec for laser runs, 5 usec for all other trigger types]
  -G, --G2 <on|off>: specifies whether to run the G2 trigger [default=on]
  -m <on|off>: specifies whether to run online monitoring [default=on]
  -D : disables the writing of data to disk
  -s <file size>: specifies the size threshold for closing data files (in MB)
      [default is 8000 MB (~7.8 GB); zero means that there is no file size limit]
  --file-events <count>: specifies the desired number of events in each file
      [default=0, which means no event count limit for files]
  --file-duration <duration>: specifies the desired duration of each file (minutes)
      [default=0, which means no duration limit for files]
  -c <compression level>: specifies the ADC data compression level
      0 = no compression
      1 = compression, both raw and compressed data kept and written to disk
      2 = compression, both sent to aggregator, only compressed written to disk
      3 = compression, only compressed sent to aggregator,
          online monitoring must decompress [default]
  -o <data dir>: specifies the directory for data files [default=/data/test]
  --veto-choice <choice>: specifies which parts of the veto system to include
      (\"off\", \"TDC\", \"ADC\", \"Hybrid\", \"TDC+ADC\", or \"TDC+Hybrid\") [default=\"TDC+ADC\"]
  --V1724-enabled <choice>: specifies whether to include the V1724 boards in the system
      (\"yes\", \"no\") [default=\"yes\"]
  --tdc <V1495_delay,V1190_gate>: a comma-separated list of two numbers that specifies
      the delay of the TDC stop pulse from the V1495 and the width of the TDC gate
      (both values in usec) [defaults are 44 usec and the width of the ADC acquisition gate]
Begin-run options (start command):
  -C <comment>: specify the run comment
  -T <type>: specify the run type
End-run options (stop command):
  -n, --run-events <count>: specifies the desired number of events in the run
      [default=0, which ends the run immediately, independent of the number of events]
  -d, --run-duration <duration>: specifies the desired length of the run (minutes)
      [default=0, which ends the run immediately, independent of the duration of the run]
Notes:
  The primary commands are the following:
   * init - initializes (configures) the DAQ processes
   * start - starts a run
   * pause - pauses the run
   * resume - resumes the run
   * stop - stops the run
   * status - checks the status of each DAQ process
   * get-legal-commands - fetches the legal commands from each DAQ process
  Additional commands include:
   * shutdown - stops the run (if one is going), resets the DAQ processes
       to their ground state (if needed), and stops the MPI program (DAQ processes)
   * start-system - starts the MPI program (the DAQ processes)
   * restart - this is the same as a shutdown followed by a start-system
   * reinit - this is the same as a shutdown followed by a start-system and an init
   * exit - this resets the DAQ processes to their ground state, stops the MPI
       program, and exits PMT.
  Expert-level commands:
   * fast-shutdown - stops the MPI program (all DAQ processes) no matter what
       state they are in. This could have bad consequences if a run is going!
   * fast-restart - this is the same as a fast-shutdown followed by a start-system
   * fast-reinit - this is the same as a fast-shutdown followed by a start-system
       and an init
   * fast-exit - this stops the MPI program, and exits PMT.
Examples: ${scriptName} -p 32768 init
" >&2
}

# parse the command-line options
originalCommand="$0 $*"
pulserRate=""
laserRate=""
randomTrigger=""
gateWidth="400"
inhibitTime=""
lowThreshold=5
compressionLevel=3
onmonEnable=on
diskWriting=1
dwChoiceSpecified=0
dataDir="/data/incoming"
runNumber=""
userSuppliedRunNumber=""
runEventCount=0
runDuration=0
fileSize=7800
fsChoiceSpecified=0
fileEventCount=0
fileDuration=0
preTriggerWindow=""
verbose=0
vetoChoice="TDC+ADC"
#vetoChoice="off"
V1724Enabled="yes"
tdcParams=""
g2Trigger=on
OPTIND=1
while getopts "hp:l:rg:G:i:c:C:T:N:o:t:m:Dn:d:s:w:v-:" opt; do
    if [ "$opt" = "-" ]; then
        opt=$OPTARG
    fi
    case $opt in
        h | help)
            usage
            exit 1
            ;;
        p)
            pulserRate=${OPTARG}
            ;;
        l)
            laserRate=${OPTARG}
            ;;
        r)
            randomTrigger=1
            ;;
        g)
            gateWidth=${OPTARG}
            ;;
        i)
            inhibitTime=${OPTARG}
            ;;
        w)
            preTriggerWindow=${OPTARG}
            ;;
        t)
            lowThreshold=${OPTARG}
            ;;
        c)
            compressionLevel=${OPTARG}
            ;;
        C)
            runComment=${OPTARG}
            ;;
        T)
            runType=${OPTARG}
            ;;
        G)
            g2Trigger=${OPTARG}
            ;;
        G2)
            g2Trigger=${!OPTIND}
            let OPTIND=$OPTIND+1
            ;;
        m)
            onmonEnable=${OPTARG}
            ;;
        N)
            userSuppliedRunNumber=${OPTARG}
            ;;
        o)
            dataDir=${OPTARG}
            ;;
        D)
            diskWriting=0
            dwChoiceSpecified=1
            ;;
        n)
            runEventCount=${OPTARG}
            ;;
        d)
            runDuration=${OPTARG}
            ;;
        run-events)
            runEventCount=${!OPTIND}
            let OPTIND=$OPTIND+1
            ;;
        run-duration)
            runDuration=${!OPTIND}
            let OPTIND=$OPTIND+1
            ;;
        file-events)
            fileEventCount=${!OPTIND}
            let OPTIND=$OPTIND+1
            ;;
        file-duration)
            fileDuration=${!OPTIND}
            let OPTIND=$OPTIND+1
            ;;
        s)
            fileSize=${OPTARG}
            fsChoiceSpecified=1
            ;;
        v)
            verbose=1
            ;;
        veto-choice)
            vetoChoice=${!OPTIND}
            let OPTIND=$OPTIND+1
            ;;
        V1724-enabled)
            V1724Enabled=${!OPTIND}
            let OPTIND=$OPTIND+1
            ;;
        tdc)
            tdcParams=${!OPTIND}
            let OPTIND=$OPTIND+1
            ;;
        *)
            usage
            exit 1
            ;;
    esac
done
shift $(($OPTIND - 1))

# fetch the command to run
if [ $# -lt 1 ]; then
    usage
    exit
fi
command=$1
shift

# verify that the command is one that we expect
if [[ "$command" != "start-system" ]] && \
   [[ "$command" != "init" ]] && \
   [[ "$command" != "start" ]] && \
   [[ "$command" != "pause" ]] && \
   [[ "$command" != "resume" ]] && \
   [[ "$command" != "stop" ]] && \
   [[ "$command" != "status" ]] && \
   [[ "$command" != "get-legal-commands" ]] && \
   [[ "$command" != "shutdown" ]] && \
   [[ "$command" != "fast-shutdown" ]] && \
   [[ "$command" != "restart" ]] && \
   [[ "$command" != "fast-restart" ]] && \
   [[ "$command" != "reinit" ]] && \
   [[ "$command" != "fast-reinit" ]] && \
   [[ "$command" != "exit" ]] && \
   [[ "$command" != "fast-exit" ]]; then
    echo "Invalid command."
    usage
    exit
fi

export PGPASSWORD=ner

# save the diskWriting choice from the init command
if [[ "$command" == "init" ]] || [[ "$command" == "reinit" ]] || \
   [[ "$command" == "fast-reinit" ]]; then
    echo "$diskWriting" > ~/.ds50daq.diskWriting
fi

# fetch the diskWriting choice for the start and stop commands
if [[ "$dwChoiceSpecified" == "0" ]]; then
    if [[ "$command" == "start" ]] || [[ "$command" == "stop" ]]; then
        if [[ -e ~/.ds50daq.diskWriting ]]; then
            diskWriting=`cat ~/.ds50daq.diskWriting`
        fi
    fi
fi

# save the runNumber choice from the start command
if [[ "$command" == "start" ]]; then
    if [[ "$userSuppliedRunNumber" == "" ]]; then
        echo "none" > ~/.ds50daq.userSuppliedRunNumber
    else
        echo "$userSuppliedRunNumber" > ~/.ds50daq.userSuppliedRunNumber
    fi
fi

# fetch the runNumber choice for the stop command
if [[ "$userSuppliedRunNumber" == "" ]]; then
    if [[ "$command" == "stop" ]]; then
        if [[ -e ~/.ds50daq.userSuppliedRunNumber ]]; then
            userSuppliedRunNumber=`cat ~/.ds50daq.userSuppliedRunNumber`
            if [[ "$userSuppliceRunNumber" == "none" ]]; then
                userSuppliedRunNumber=""
            fi
        fi
    fi
fi

# verify that the expected arguments were provided
if [[ "$command" == "start" ]]; then
  if [[ "$diskWriting" == "1" ]]; then 
    if [[ "$runComment" == "" ]] && [[ "$runType" == "" ]]; then
      echo ""
      echo "*** RunType and RunComment must be specified"
      usage
      exit
    fi
    end=`psql -A -c 'select endtime from run.daqruns_old_format order by runid desc limit 1' -h dsdb.lngs.infn.it darkside50 run | grep ^2`
    if [[ "$end" == "" ]]; then
      echo "Previous run not stopped"
      exit
    fi
  fi
  if [[ "$runNumber" == "" ]]; then
    if [[ "$userSuppliedRunNumber" != "" ]]; then
      runNumber=${userSuppliedRunNumber}
    else
      if [[ "$diskWriting" == "1" ]]; then 
        runNumber=`psql -A -c 'select MAX(runid)+1 from run.daqruns_old_format' -h dsdb.lngs.infn.it darkside50 run | grep ^[0-9]`
      fi
    fi
  fi
fi

if [[ "$command" == "stop" ]]; then
  if [[ "$diskWriting" == "1" ]]; then 
    if [[ "$userSuppliedRunNumber" == "" ]]; then
      end=`psql -A -c 'select endtime from run.daqruns_old_format order by runid desc limit 1' -h dsdb.lngs.infn.it darkside50 run | grep ^2`
      if [[ "$end" != "" ]]; then
        echo "There is not an active run"
        exit
      fi
    fi
  fi
fi

# fill in values for options that weren't specified
if [[ "$pulserRate" == "" ]]; then
    pulserRate=0
fi
if [[ "$laserRate" == "" ]]; then
    laserRate=0
fi
let laserRate=$laserRate+0
if [[ "$randomTrigger" == "" ]]; then
    randomTrigger=0
fi
if [[ "$runNumber" == "" ]]; then
    runNumber=100
fi
if [[ "$preTriggerWindow" == "" ]]; then
    if [[ $laserRate -ne 0 ]]; then
        preTriggerWindow=1
    else
        preTriggerWindow=5
    fi
fi
if [[ "$inhibitTime" == "" ]]; then
    let inhibitTimeVal=${gateWidth}+0
    if [[ $inhibitTimeVal > 150 ]]; then
        let inhibitTimeVal=${inhibitTimeVal}-60
        let inhibitTimeVal=${inhibitTimeVal}*2
        let inhibitTimeVal=${inhibitTimeVal}+50
    fi
    inhibitTime="$inhibitTimeVal"
fi
# if no TDC parameters were specified, use 44 usec and the ADC gate width
if [[ "$tdcParams" == "" ]]; then
    tdcParams="44,${gateWidth}"
fi
# if the TDC parameter list was specified and one of the two values
# was left blank, fill in the blank value with the default value
tdcParams=`echo ${tdcParams} | sed "s/^,/44,/"`
tdcParams=`echo ${tdcParams} | sed "s/,$/,${gateWidth}/"`
# if the TDC parameter list consists of a single number, use that
# number for both the TDC delay and the TDC gate width
if [[ `echo ${tdcParams} | sed 's/,/ /' | wc -w` -eq 1 ]]; then
    tdcParams="${tdcParams},${tdcParams}"
fi

# reduce the default fileSize threshold if the gate width is large
if [[ "$fsChoiceSpecified" == "0" ]]; then
    let gateWidthVal=${gateWidth}+0
    if [[ ${gateWidthVal} -ge 900 ]]; then
        fileSize=5000
    elif [[ ${gateWidthVal} -ge 700 ]]; then
        fileSize=5500
    elif [[ ${gateWidthVal} -ge 500 ]]; then
        fileSize=6000
    elif [[ ${gateWidthVal} -ge 300 ]]; then
        fileSize=6500
    elif [[ ${gateWidthVal} -ge 100 ]]; then
        fileSize=7000
    fi
fi

# translate the G2 trigger flag
if [[ "$g2Trigger" == "on" ]]; then
    g2Trigger=1
else
    g2Trigger=0
fi

# translate the onmon enable flag
if [[ "$onmonEnable" == "on" ]]; then
    onmonEnable=1
else
    onmonEnable=0
fi

# parse out the TDC delay and gate width
tdcDelay=`echo ${tdcParams} | sed 's/,/ /' | cut -f1 -d' '`
tdcGateWidth=`echo ${tdcParams} | sed 's/,/ /' | cut -f2 -d' '`

# verify that we don't have both a number of events and a time limit
# for stopping a run
if [[ "$command" == "stop" ]]; then
    if [[ "$runEventCount" != "0" ]] && [[ "$runDuration" != "0" ]]; then
        echo ""
        echo "*** Only one of event count or run duration may be specified."
        usage
        exit
    fi
fi

# build the logfile name
TIMESTAMP=`date '+%Y%m%d%H%M%S'`
logFile="/daqlogs/masterControl/dsMC-${TIMESTAMP}-${command}.log"
echo "${originalCommand}" > $logFile
echo ">>> ${originalCommand} (Disk writing is ${diskWriting})"

# calculate the shmkey that should be checked
let shmKey=1078394880+${pmtPort}
shmKeyString=`printf "0x%x" ${shmKey}`

# invoke the requested command
if [[ "$command" == "shutdown" ]]; then
    # first send a stop command to end the run (in case it is needed)
    launch "stop" $runNumber $pulserRate $laserRate $randomTrigger \
        $gateWidth $lowThreshold $compressionLevel $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $preTriggerWindow $fileEventCount $fileDuration $verbose $vetoChoice \
        $g2Trigger $V1724Enabled $inhibitTime $tdcDelay $tdcGateWidth
    # next send a shutdown command to move the processes to their ground state
    launch "shutdown" $runNumber $pulserRate $laserRate $randomTrigger \
        $gateWidth $lowThreshold $compressionLevel $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $preTriggerWindow $fileEventCount $fileDuration $verbose $vetoChoice \
        $g2Trigger $V1724Enabled $inhibitTime $tdcDelay $tdcGateWidth
    # stop the MPI program
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.stopSystem
    # clean up any stale shared memory segment
    rsh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"
elif [[ "$command" == "start-system" ]]; then
    rsh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.startSystem
elif [[ "$command" == "restart" ]]; then
    # first send a stop command to end the run (in case it is needed)
    launch "stop" $runNumber $pulserRate $laserRate $randomTrigger \
        $gateWidth $lowThreshold $compressionLevel $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $preTriggerWindow $fileEventCount $fileDuration $verbose $vetoChoice \
        $g2Trigger $V1724Enabled $inhibitTime $tdcDelay $tdcGateWidth
    # next send a shutdown command to move the processes to their ground state
    launch "shutdown" $runNumber $pulserRate $laserRate $randomTrigger \
        $gateWidth $lowThreshold $compressionLevel $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $preTriggerWindow $fileEventCount $fileDuration $verbose $vetoChoice \
        $g2Trigger $V1724Enabled $inhibitTime $tdcDelay $tdcGateWidth
    # stop the MPI program
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.stopSystem
    # clean up any stale shared memory segment
    rsh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"
    # start the MPI program
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.startSystem
elif [[ "$command" == "reinit" ]]; then
    # first send a stop command to end the run (in case it is needed)
    launch "stop" $runNumber $pulserRate $laserRate $randomTrigger \
        $gateWidth $lowThreshold $compressionLevel $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $preTriggerWindow $fileEventCount $fileDuration $verbose $vetoChoice \
        $g2Trigger $V1724Enabled $inhibitTime $tdcDelay $tdcGateWidth
    # next send a shutdown command to move the processes to their ground state
    launch "shutdown" $runNumber $pulserRate $laserRate $randomTrigger \
        $gateWidth $lowThreshold $compressionLevel $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $preTriggerWindow $fileEventCount $fileDuration $verbose $vetoChoice \
        $g2Trigger $V1724Enabled $inhibitTime $tdcDelay $tdcGateWidth
    # stop the MPI program
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.stopSystem
    # clean up any stale shared memory segment
    rsh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"
    # start the MPI program
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.startSystem
    # send the init command to re-initialize the system
    sleep 5
    launch "init" $runNumber $pulserRate $laserRate $randomTrigger \
        $gateWidth $lowThreshold $compressionLevel $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $preTriggerWindow $fileEventCount $fileDuration $verbose $vetoChoice \
        $g2Trigger $V1724Enabled $inhibitTime $tdcDelay $tdcGateWidth
elif [[ "$command" == "exit" ]]; then
    launch "shutdown" $runNumber $pulserRate $laserRate $randomTrigger \
        $gateWidth $lowThreshold $compressionLevel $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $preTriggerWindow $fileEventCount $fileDuration $verbose $vetoChoice \
        $g2Trigger $V1724Enabled $inhibitTime $tdcDelay $tdcGateWidth
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.stopSystem
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.exit
    rsh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"

elif [[ "$command" == "fast-shutdown" ]]; then
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.stopSystem
    rsh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"
elif [[ "$command" == "fast-restart" ]]; then
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.stopSystem
    rsh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.startSystem
elif [[ "$command" == "fast-reinit" ]]; then
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.stopSystem
    rsh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.startSystem
    sleep 5
    launch "init" $runNumber $pulserRate $laserRate $randomTrigger \
        $gateWidth $lowThreshold $compressionLevel $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $preTriggerWindow $fileEventCount $fileDuration $verbose $vetoChoice \
        $g2Trigger $V1724Enabled $inhibitTime $tdcDelay $tdcGateWidth
elif [[ "$command" == "fast-exit" ]]; then
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.stopSystem
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.exit
    rsh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"

else
    launch $command $runNumber $pulserRate $laserRate $randomTrigger \
        $gateWidth $lowThreshold $compressionLevel $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $preTriggerWindow $fileEventCount $fileDuration $verbose $vetoChoice \
        $g2Trigger $V1724Enabled $inhibitTime $tdcDelay $tdcGateWidth
fi

if [[ "$command" == "start" ]]; then
  if [[ "$diskWriting" == "1" ]]; then 
    if [[ "$userSuppliedRunNumber" == "" ]]; then
      psql -q -c "insert INTO run.daqruns_old_format (runid, starttime, type, comment) VALUES ($runNumber, 'NOW', '$runType', '$runComment');"  -h dsdb.lngs.infn.it darkside50 run
    fi
  fi 
  echo "Run $runNumber started"
  echo "Run $runNumber started" >> $logFile
fi

if [[ "$command" == "stop" ]]; then
  if [[ "$diskWriting" == "1" ]]; then 
    if [[ "$userSuppliedRunNumber" == "" ]]; then
      psql -q -c "UPDATE run.daqruns_old_format SET endtime='NOW' where runid=(select MAX(runid) from run.daqruns_old_format);"  -h dsdb.lngs.infn.it darkside50 run 
    fi
  fi
fi
