#!/bin/bash

source `which setPortNumbers.sh`

AGGREGATOR_NODE="dsfr6"
THIS_NODE=`hostname -s`

# this function expects a number of arguments:
#  1) the DAQ command to be sent
#  2) the run number (dummy for non-start commands)
#  3) the pulser trigger rate
#  4) the laser trigger rate
#  5) whether to enable the random trigger (0 or 1)
#  6) the acqusition gate width (usec)
#  7) the majority logic low threshold
#  8) the compression level for ADC data [0..2]
#  9) whether to run online monitoring
# 10) the data directory
# 11) the logfile name
# 12) whether to write data to disk [0,1]
# 13) the desired number of events in the run (stop command)
# 14) the desired duration of the run (minutes, stop command)
# 15) the desired size of each data file
# 16) the pre-trigger window size (usec)
# 17) the desired number of events in each file
# 18) the desired time duration of each file (minutes)
# 19) whether to print out CFG information (verbose)
# 20) the inhibit time (usec)
function launch() {
  let ebComp=$8+10
  let dwAgComp=$8+10
  let onmonAgComp=$8+10
  if [[ "$8" == "3" ]]; then
    ebComp=12
    dwAgComp=10
    onmonAgComp=14
  elif [[ "$8" == "2" ]]; then
    ebComp=11
    dwAgComp=13
    onmonAgComp=10
  elif [[ "$8" == "1" ]]; then
    ebComp=11
    dwAgComp=10
    onmonAgComp=10
  fi
  enableSerial=""
  if [[ "${19}" == "1" ]]; then
      enableSerial="-e"
  fi

  ds50MasterControl.rb ${enableSerial} -s -c $1 \
    --v1495 dsfr6,$frPort0,0,$3,$4,$5,$6,$7,1,0,0,${20},$6 \
    --v1720 dsfr6,$frPort1,1,0,$6,${16},0 \
    --v1720 dsfr6,$frPort2,2,0,$6,${16},1 \
    --eb dseb8,$ebPort0,$ebComp \
    --eb dseb8,$ebPort1,$ebComp \
    --eb dseb8,$ebPort2,$ebComp \
    --eb dseb8,$ebPort3,$ebComp \
    --eb dseb8,$ebPort4,$ebComp \
    --ag ${AGGREGATOR_NODE},$agPort0,1,$dwAgComp \
    --ag ${AGGREGATOR_NODE},$agPort1,1,$onmonAgComp \
    --data-dir ${10} --online-monitoring $9 \
    --write-data ${12} --run-event-count ${13} \
    --run-duration ${14} --file-size ${15} \
    --file-event-count ${17} --file-duration ${18} \
    --compression-table-from-db 0 \
    --run-number $2 2>&1 | tee -a ${11}
}

scriptName=`basename $0`
usage () {
    echo "
Usage: ${scriptName} [options] <command>
Where command is one of:
  init, start, pause, resume, stop, status, get-legal-commands,
  shutdown, start-system, restart, reinit, exit,
  fast-shutdown, fast-restart, fast-reinit, or fast-exit
General options:
  -h, --help: prints this usage message
Configuration options (init commands):
  -t <low threshold>: specifies the low threshold for the majority logic
      trigger [default=5; zero means that majority logic trigger is off]
  -l <laser rate>: specifies the rate of the laser trigger
      [default=0, which means that the laser trigger is off]
  -p <pulser rate>: specifies the rate of the pulser trigger
      [default=0, which means that the pulser trigger is off]
  -r : enables the random trigger [default is random trigger off]
  -g <gate width>: specifies the acquisition gate width in usec [default=400]
  -i <inhibit time>: specifies the inhibit time in usec
      [default is a function of the gate width]
  -w <pre-trigger window>: specifies the size of the pre-trigger window in usec
      [default is 1 usec for laser runs, 5 usec for all other trigger types]
  -m <on|off>: specifies whether to run online monitoring [default=off]
  -D : disables the writing of data to disk
  -s <file size>: specifies the size threshold for closing data files (in MB)
      [default is 8000 MB (~7.8 GB); zero means that there is no file size limit]
  --file-events <count>: specifies the desired number of events in each file
      [default=0, which means no event count limit for files]
  --file-duration <duration>: specifies the desired duration of each file (minutes)
      [default=0, which means no duration limit for files]
  -c <compression level>: specifies the ADC data compression level
      0 = no compression [default]
      1 = compression, both raw and compressed data kept and written to disk
      2 = compression, both sent to aggregator, only compressed written to disk
      3 = compression, compressed sent to aggregator, online monitoring must decompress
  -o <data dir>: specifies the directory for data files [default=/data/test]
Begin-run options (start command):
  -N <run number>: specifies the run number
End-run options (stop command):
  -n, --run-events <count>: specifies the desired number of events in the run
      [default=0, which ends the run immediately, independent of the number of events]
  -d, --run-duration <duration>: specifies the desired length of the run (minutes)
      [default=0, which ends the run immediately, independent of the duration of the run]
Notes:
  The start command expects a run number to be specified (-N).
  The primary commands are the following:
   * init - initializes (configures) the DAQ processes
   * start - starts a run
   * pause - pauses the run
   * resume - resumes the run
   * stop - stops the run
   * status - checks the status of each DAQ process
   * get-legal-commands - fetches the legal commands from each DAQ process
  Additional commands include:
   * shutdown - stops the run (if one is going), resets the DAQ processes
       to their ground state (if needed), and stops the MPI program (DAQ processes)
   * start-system - starts the MPI program (the DAQ processes)
   * restart - this is the same as a shutdown followed by a start-system
   * reinit - this is the same as a shutdown followed by a start-system and an init
   * exit - this resets the DAQ processes to their ground state, stops the MPI
       program, and exits PMT.
  Expert-level commands:
   * fast-shutdown - stops the MPI program (all DAQ processes) no matter what
       state they are in. This could have bad consequences if a run is going!
   * fast-restart - this is the same as a fast-shutdown followed by a start-system
   * fast-reinit - this is the same as a fast-shutdown followed by a start-system
       and an init
   * fast-exit - this stops the MPI program, and exits PMT.
Examples: ${scriptName} -p 32768 init
          ${scriptName} -N 101 start
" >&2
}

# parse the command-line options
originalCommand="$0 $*"
pulserRate=""
laserRate=""
randomTrigger=""
gateWidth="400"
inhibitTime=""
lowThreshold=5
compressionLevel=0
onmonEnable=off
diskWriting=1
dataDir="/data/test"
runNumber=""
runEventCount=0
runDuration=0
fileSize=8000
fsChoiceSpecified=0
fileEventCount=0
fileDuration=0
preTriggerWindow=""
verbose=0
OPTIND=1
while getopts "hp:l:rg:i:c:N:o:t:m:Dn:d:s:w:v-:" opt; do
    if [ "$opt" = "-" ]; then
        opt=$OPTARG
    fi
    case $opt in
        h | help)
            usage
            exit 1
            ;;
        p)
            pulserRate=${OPTARG}
            ;;
        l)
            laserRate=${OPTARG}
            ;;
        r)
            randomTrigger=1
            ;;
        g)
            gateWidth=${OPTARG}
            ;;
        i)
            inhibitTime=${OPTARG}
            ;;
        w)
            preTriggerWindow=${OPTARG}
            ;;
        t)
            lowThreshold=${OPTARG}
            ;;
        c)
            compressionLevel=${OPTARG}
            ;;
        N)
            runNumber=${OPTARG}
            ;;
        m)
            onmonEnable=${OPTARG}
            ;;
        o)
            dataDir=${OPTARG}
            ;;
        D)
            diskWriting=0
            ;;
        n)
            runEventCount=${OPTARG}
            ;;
        d)
            runDuration=${OPTARG}
            ;;
        run-events)
            runEventCount=${!OPTIND}
            let OPTIND=$OPTIND+1
            ;;
        run-duration)
            runDuration=${!OPTIND}
            let OPTIND=$OPTIND+1
            ;;
        file-events)
            fileEventCount=${!OPTIND}
            let OPTIND=$OPTIND+1
            ;;
        file-duration)
            fileDuration=${!OPTIND}
            let OPTIND=$OPTIND+1
            ;;
        s)
            fileSize=${OPTARG}
            fsChoiceSpecified=1
            ;;
        v)
            verbose=1
            ;;
        *)
            usage
            exit 1
            ;;
    esac
done
shift $(($OPTIND - 1))

# fetch the command to run
if [ $# -lt 1 ]; then
    usage
    exit
fi
command=$1
shift

# verify that the command is one that we expect
if [[ "$command" != "start-system" ]] && \
   [[ "$command" != "init" ]] && \
   [[ "$command" != "start" ]] && \
   [[ "$command" != "pause" ]] && \
   [[ "$command" != "resume" ]] && \
   [[ "$command" != "stop" ]] && \
   [[ "$command" != "status" ]] && \
   [[ "$command" != "get-legal-commands" ]] && \
   [[ "$command" != "shutdown" ]] && \
   [[ "$command" != "fast-shutdown" ]] && \
   [[ "$command" != "restart" ]] && \
   [[ "$command" != "fast-restart" ]] && \
   [[ "$command" != "reinit" ]] && \
   [[ "$command" != "fast-reinit" ]] && \
   [[ "$command" != "exit" ]] && \
   [[ "$command" != "fast-exit" ]]; then
    echo "Invalid command."
    usage
    exit
fi

# verify that the expected arguments were provided
if [[ "$command" == "start" ]] && [[ "$runNumber" == "" ]]; then
    echo ""
    echo "*** A run number needs to be specified."
    usage
    exit
fi

# fill in values for options that weren't specified
if [[ "$pulserRate" == "" ]]; then
    pulserRate=0
fi
if [[ "$laserRate" == "" ]]; then
    laserRate=0
fi
let laserRate=$laserRate+0
if [[ "$randomTrigger" == "" ]]; then
    randomTrigger=0
fi
if [[ "$runNumber" == "" ]]; then
    runNumber=101
fi
if [[ "$preTriggerWindow" == "" ]]; then
    if [[ $laserRate -ne 0 ]]; then
        preTriggerWindow=1
    else
        preTriggerWindow=5
    fi
fi
if [[ "$inhibitTime" == "" ]]; then
    let inhibitTimeVal=${gateWidth}+0
    if [[ $inhibitTimeVal > 150 ]]; then
        let inhibitTimeVal=${inhibitTimeVal}-60
        let inhibitTimeVal=${inhibitTimeVal}*2
        let inhibitTimeVal=${inhibitTimeVal}+50
    fi
    inhibitTime="$inhibitTimeVal"
fi

# reduce the default fileSize threshold if the gate width is large
if [[ "$fsChoiceSpecified" == "0" ]]; then
    let gateWidthVal=${gateWidth}+0
    if [[ ${gateWidthVal} -ge 300 ]]; then
        fileSize=6000
    elif [[ ${gateWidthVal} -ge 100 ]]; then
        fileSize=7000
    fi
fi

# translate the onmon enable flag
if [[ "$onmonEnable" == "on" ]]; then
    onmonEnable=1
else
    onmonEnable=0
fi

# verify that we don't have both a number of events and a time limit
# for stopping a run
if [[ "$command" == "stop" ]]; then
    if [[ "$runEventCount" != "0" ]] && [[ "$runDuration" != "0" ]]; then
        echo ""
        echo "*** Only one of event count or run duration may be specified."
        usage
        exit
    fi
fi

# build the logfile name
TIMESTAMP=`date '+%Y%m%d%H%M%S'`
logFile="/daqlogs/masterControl/dsMC-${TIMESTAMP}-${command}.log"
echo "${originalCommand}" > $logFile
echo ">>> ${originalCommand} (Disk writing is ${diskWriting})"

# calculate the shmkey that should be checked
let shmKey=1078394880+${pmtPort}
shmKeyString=`printf "0x%x" ${shmKey}`

# invoke the requested command
if [[ "$command" == "shutdown" ]]; then
    # first send a stop command to end the run (in case it is needed)
    launch "stop" $runNumber $pulserRate $laserRate $randomTrigger \
        $gateWidth $lowThreshold $compressionLevel $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $preTriggerWindow $fileEventCount $fileDuration $verbose $inhibitTime
    # next send a shutdown command to move the processes to their ground state
    launch "shutdown" $runNumber $pulserRate $laserRate $randomTrigger \
        $gateWidth $lowThreshold $compressionLevel $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $preTriggerWindow $fileEventCount $fileDuration $verbose $inhibitTime
    # stop the MPI program
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.stopSystem
    # clean up any stale shared memory segment
    rsh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"
elif [[ "$command" == "start-system" ]]; then
    rsh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.startSystem
elif [[ "$command" == "restart" ]]; then
    # first send a stop command to end the run (in case it is needed)
    launch "stop" $runNumber $pulserRate $laserRate $randomTrigger \
        $gateWidth $lowThreshold $compressionLevel $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $preTriggerWindow $fileEventCount $fileDuration $verbose $inhibitTime
    # next send a shutdown command to move the processes to their ground state
    launch "shutdown" $runNumber $pulserRate $laserRate $randomTrigger \
        $gateWidth $lowThreshold $compressionLevel $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $preTriggerWindow $fileEventCount $fileDuration $verbose $inhibitTime
    # stop the MPI program
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.stopSystem
    # clean up any stale shared memory segment
    rsh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"
    # start the MPI program
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.startSystem
elif [[ "$command" == "reinit" ]]; then
    # first send a stop command to end the run (in case it is needed)
    launch "stop" $runNumber $pulserRate $laserRate $randomTrigger \
        $gateWidth $lowThreshold $compressionLevel $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $preTriggerWindow $fileEventCount $fileDuration $verbose $inhibitTime
    # next send a shutdown command to move the processes to their ground state
    launch "shutdown" $runNumber $pulserRate $laserRate $randomTrigger \
        $gateWidth $lowThreshold $compressionLevel $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $preTriggerWindow $fileEventCount $fileDuration $verbose $inhibitTime
    # stop the MPI program
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.stopSystem
    # clean up any stale shared memory segment
    rsh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"
    # start the MPI program
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.startSystem
    # send the init command to re-initialize the system
    sleep 5
    launch "init" $runNumber $pulserRate $laserRate $randomTrigger \
        $gateWidth $lowThreshold $compressionLevel $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $preTriggerWindow $fileEventCount $fileDuration $verbose $inhibitTime
elif [[ "$command" == "exit" ]]; then
    launch "shutdown" $runNumber $pulserRate $laserRate $randomTrigger \
        $gateWidth $lowThreshold $compressionLevel $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $preTriggerWindow $fileEventCount $fileDuration $verbose $inhibitTime
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.stopSystem
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.exit
    rsh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"

elif [[ "$command" == "fast-shutdown" ]]; then
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.stopSystem
    rsh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"
elif [[ "$command" == "fast-restart" ]]; then
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.stopSystem
    rsh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.startSystem
elif [[ "$command" == "fast-reinit" ]]; then
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.stopSystem
    rsh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.startSystem
    sleep 5
    launch "init" $runNumber $pulserRate $laserRate $randomTrigger \
        $gateWidth $lowThreshold $compressionLevel $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $preTriggerWindow $fileEventCount $fileDuration $verbose $inhibitTime
elif [[ "$command" == "fast-exit" ]]; then
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.stopSystem
    xmlrpc ${THIS_NODE}:${pmtPort}/RPC2 pmt.exit
    rsh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"

else
    launch $command $runNumber $pulserRate $laserRate $randomTrigger \
        $gateWidth $lowThreshold $compressionLevel $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $preTriggerWindow $fileEventCount $fileDuration $verbose $inhibitTime
fi
