#include "ds50daq/DAQ/V1190Driver.hh"

#include "artdaq/DAQdata/Debug.hh"
#include "artdaq/Application/GeneratorMacros.hh"
#include "cetlib/exception.h"
#include "darksidecore/Data/Config.hh"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "detail/V1190_io.hh"

#include <string>
#include <string.h>
#include <vector>

using fhicl::ParameterSet;
using ds50::V1190Fragment;
namespace V1190 = detail::V1190;

ds50::V1190Driver::V1190Driver(ParameterSet const & ps): DS50FragmentGenerator(ps.get<ParameterSet> ("generator_ds50")) {
  init_link (ps);

  init_metadata (ps);

  init_trigger (ps);

  init_channels (ps);
}

ds50::V1190Driver::~V1190Driver() noexcept {
  close_link ();
}

void ds50::V1190Driver::close_link () {
  v1190_ = 0;
}


void ds50::V1190Driver::init_link (ParameterSet const& ps) {
  v1190_ = std::unique_ptr<detail::V1190_io>(new detail::V1190_io (ps.get<uint32_t> ("vme_base"), ps.get<std::string> ("link_type"), ps.get<int> ("link_number"), ps.get<int> ("link_node", 0), ps.get<uint8_t> ("irq_level", 2), ps.get<uint8_t> ("irq_vector", 90)));

  v1190_->termination (ps.get<bool> ("control_lines_termination", true));
  v1190_->empty_events (true);

  io_buffer_length_ = ps.get<int> ("io_buffer_length", 40000);
  io_buffer_ = std::unique_ptr<V1190Fragment::Word[]>(new V1190Fragment::Word[io_buffer_length_]);

  v1190_->geo (board_id ());

  v1190_->out_prog (20000); // max is 32k

  v1190_->blt_ro (true, 200);
}

void ds50::V1190Driver::init_metadata (ParameterSet const& ps) {
  metadata_.serial0 = v1190_->serial0 (); 
  metadata_.serial1 = v1190_->serial1 (); 
  metadata_.firmware_version = v1190_->fw_release ();
  metadata_.gate_ns = std::min(ps.get<float> ("acquisition_gate_us"), 50.F) * 1000;
  metadata_.pre_trigger_ns = std::min(ps.get<float> ("post_trigger_us"), 0.95F) * 1000 - metadata_.gate_ns;
  //metadata_.gate_ns = 4000 + 20000;
  //metadata_.pre_trigger_ns = 500 - metadata_.gate_ns;
  metadata_.resolution_ps = 800;

  mf::LogInfo("V1190") << "board_id " << board_id () << " found TDC unit with serial " << std::hex << std::showbase <<  metadata_.serial0 << ":" << metadata_.serial1 << " fw release " << metadata_.firmware_version;
}

void ds50::V1190Driver::init_trigger (ParameterSet const& ps) {
  v1190_->acquisition_mode (V1190::trigger_matching);
  v1190_->edge_detection (V1190::trailing_and_leading);
  v1190_->edge_detection (V1190::leading_only);
  v1190_->edge_resolution (V1190::ps800);
  //v1190_->channel_dead_time (V1190::cdt_30ns);
  v1190_->channel_dead_time (V1190::cdt_5ns);

  v1190_->matching_window_width_25ns (int(metadata_.gate_ns / 25));

  v1190_->matching_window_offset_25ns (int16_t(metadata_.pre_trigger_ns / 25));

  v1190_->trigger_time_tag (ps.get<bool> ("trigger_time_tag", false));
  //v1190_->tdc_header (ps.get<bool> ("tdc_header", false));
  v1190_->trigget_time_subtraction (ps.get<bool> ("trigget_time_subtraction", true));
}

void ds50::V1190Driver::init_channels (ParameterSet const&) {
}

void ds50::V1190Driver::start_ () {
  v1190_->clear ();
}

void ds50::V1190Driver::stop_ () {
}


bool ds50::V1190Driver::getNext__(artdaq::FragmentPtrs& frags) {
  bool will_stop = should_stop ();

  //v1190_->sw_trigger ();
  while (1) {
    uint16_t status = v1190_->status ();
    if (status & detail::V1190::trigger_lost_msk) throw cet::exception("V1190DriverGenerator") << "board_id " << board_id () << " trigger lost at event " << ev_counter ();
    if (status & detail::V1190::error_any_msk) throw cet::exception("V1190DriverGenerator") << "board_id " << board_id () << " error detected by TDC at event " << ev_counter ();
    if (status & detail::V1190::full_msk) throw cet::exception("V1190DriverGenerator") << "board_id " << board_id () << " full at event " << ev_counter ();
    if (status & detail::V1190::almost_full_msk) mf::LogError ("V1190") << "board_id " << board_id () << " almost full at event " << ev_counter () << std::hex << std::showbase << " with status " << status;
    if (~status & detail::V1190::dready_msk) break;

    ssize_t ro_size = v1190_->read_event (io_buffer_.get (), io_buffer_length_ * sizeof (io_buffer_[0]));
    if (ro_size == 0) continue;
    else if (ro_size < 0) ro_size = -ro_size;

    size_t current_start = 0;
    while (1) {
      if ((current_start * sizeof (io_buffer_[0])) >= size_t(ro_size)) break;

      V1190Fragment::Word header(io_buffer_.get ()[current_start]);
      if (header.type () == ds50::V1190Fragment::Word::filler) {
	current_start ++;
	continue;
      }
      if (header.type () != ds50::V1190Fragment::Word::global_header) throw cet::exception("V1190DriverGenerator") << "board_id " << board_id () << " first word is not a global header " << int(header.type ()) << " " << current_start;

      size_t event_lenght = 1;
      while (1) {
	if ((current_start + event_lenght) * sizeof (io_buffer_[0]) >= size_t(ro_size)) break;
	if (io_buffer_.get ()[current_start + event_lenght++].type () == ds50::V1190Fragment::Word::trailer) break;
      }
      V1190Fragment::Word trailer(io_buffer_.get ()[current_start + event_lenght - 1]);
      if (trailer.type () != ds50::V1190Fragment::Word::trailer) throw cet::exception("V1190DriverGenerator") << "board_id " << board_id () << " end is not a trailer" << int(trailer.type ()) << " " << current_start;
      if (event_lenght != trailer.word_count ()) throw cet::exception("V1190DriverGenerator") << "board_id " << board_id () << " event size mismatch " << event_lenght << " vs " << trailer.word_count () ;

      std::unique_ptr<artdaq::Fragment> frag(new artdaq::Fragment(std::ceil (event_lenght * sizeof (io_buffer_[0]) / static_cast<double>(sizeof (artdaq::RawDataType))), ev_counter (), fragment_id (), Config::V1190_FRAGMENT_TYPE, metadata_));
      memcpy (&*frag->dataBegin (), io_buffer_.get () + current_start, event_lenght * sizeof (io_buffer_[0]));
      V1190Fragment const v1190_frag(*frag.get ());

      //std::cout << std::hex << std::showbase << v1190_frag << std::endl;
      //std::cout << board_id () << " Ev " << ev_counter () << " " << v1190_frag.dataBegin ()[0].event_count () << std::endl;

      int ev_counter_diff = v1190_frag.dataBegin ()[0].event_count () + 1 - ev_counter ();
      if (ev_counter_diff > 0) {
	mf::LogError ("V1190") << "board_id " << board_id () << " jump " << ev_counter_diff << "(got " << v1190_frag.dataBegin ()[0].event_count () << ") at event " << ev_counter ();

	for (int i = 0; ev_counter_diff < 100 && i < ev_counter_diff; ++i)
	  frags.emplace_back(new artdaq::Fragment(ev_counter () + i, fragment_id (), Config::MISSED_FRAGMENT_TYPE));

	ev_counter_inc (ev_counter_diff);
	frag->setSequenceID (ev_counter ());
      } else if (ev_counter_diff < 0) {
	mf::LogError ("V1190") << "board_id " << board_id () << " duplicated fragment " << v1190_frag.dataBegin ()[0].event_count () << " at event " << ev_counter ();
	continue;
      }

      frags.emplace_back (std::move (frag));

      ev_counter_inc ();
      current_start += event_lenght;
    }

    //if (frags.size () > 10) break;
  }

  if (!frags.size () && will_stop) return false;
  return true;
}

DEFINE_ARTDAQ_COMMANDABLE_GENERATOR(ds50::V1190Driver)
